#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Utilities for private pip space to download packages into."""

import os
import re
import subprocess
import sys

from nuitka.__past__ import total_ordering
from nuitka.containers.OrderedDicts import OrderedDict
from nuitka.PythonVersions import (
    getSitePackageCandidateNames,
    getSystemPrefixPath,
    python_version_str,
)

from .Download import getDownloadCacheDir, shouldDownload
from .Execution import (
    NuitkaCalledProcessError,
    check_output,
    getExecutablePath,
    getNullOutput,
    withEnvironmentPathAdded,
)
from .FileOperations import (
    areSamePaths,
    getFileContentByLine,
    getNormalizedPathJoin,
)
from .Hashing import HashCRC32
from .Importing import withTemporarySysPathExtension
from .Utils import getArchitecture, getOS, isWin32Windows


def getSystemPrefixExecutable():
    """Return the sys.executable of the system prefix, i.e. breaking out of virtualenv.

    See getSystemPrefixPath() for details.

    Returns:
        str - path to system executable
    """
    # Yes, we want and need it here, pylint: disable=protected-access
    if hasattr(sys, "_base_executable"):
        return sys._base_executable

    sys_prefix = getSystemPrefixPath()
    if sys_prefix == sys.prefix:
        return sys.executable

    if areSamePaths(sys_prefix, sys.prefix):
        return sys.executable

    # The "sys.executable" is a good guess for the name, but we might have to
    # look for "python" effectively.
    candidate_names = (
        os.path.basename(sys.executable),
        "python.exe" if os.name == "nt" else "python",
        "python%s" % python_version_str,
    )

    for candidate_name in candidate_names:
        if os.name == "nt":
            candidate = os.path.join(sys_prefix, candidate_name)
        else:
            candidate = os.path.join(sys_prefix, "bin", candidate_name)

        if os.path.exists(candidate):
            return candidate

    # Fallback to sys.executable, which is not correct if we are in a virtualenv
    # but the best we can do if we didn't find the system prefix one.
    return sys.executable


def _getCandidateBinPaths(logger, site_packages):
    """Get the candidate binary paths for the private pip space."""
    download_folder = getPrivatePipBaseFolder()

    candidate_bin_paths = [
        os.path.join(download_folder, "bin"),
        os.path.join(download_folder, "usr", "bin"),
        os.path.join(download_folder, "usr", "local", "bin"),
    ]

    if isWin32Windows():
        candidate_bin_paths.insert(0, os.path.join(download_folder, "Scripts"))

    if site_packages is None:
        site_packages = getPrivatePipSitePackagesDir(logger=logger)

    if site_packages:
        if isWin32Windows():
            candidate_bin_paths.insert(
                0,
                os.path.join(
                    os.path.dirname(os.path.dirname(site_packages)), "Scripts"
                ),
            )
        else:
            candidate_bin_paths.insert(
                0,
                os.path.join(
                    os.path.dirname(os.path.dirname(os.path.dirname(site_packages))),
                    "bin",
                ),
            )

    return candidate_bin_paths


def _getPrivatePipBinaryPath(
    logger,
    binary_name,
    package_name,
    module_name,
    version,
    assume_yes_for_downloads,
    reject_message,
):
    """Get the path of a binary from the private pip space."""
    candidate_bin_paths = _getCandidateBinPaths(logger=logger, site_packages=None)

    # Construct an extra_dir for search
    extra_dir = os.pathsep.join(candidate_bin_paths)

    binary_path = getExecutablePath(binary_name, extra_dir=extra_dir)

    if binary_path is not None:
        if version is None:
            return binary_path, None, assume_yes_for_downloads

        with withPrivatePipSitePackagesPathAdded(logger=logger):
            ok, _message = _checkRequiredVersion(
                logger=logger, tool=binary_name, tool_call=[binary_path]
            )
        if ok:
            return binary_path, None, assume_yes_for_downloads

    # Download, avoiding to use the result, which is the site-packages
    # folder
    site_packages_folder, assume_yes_for_downloads = tryDownloadPackageName(
        logger=logger,
        package_name=package_name,
        module_name=module_name,
        package_version=version,
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message=reject_message,
    )

    if site_packages_folder:
        # Check standard locations in the downloaded environment.
        for candidate in _getCandidateBinPaths(
            logger=logger, site_packages=site_packages_folder
        ):
            possible = os.path.join(candidate, binary_name)
            if isWin32Windows():
                possible += ".exe"

            if os.path.exists(possible):
                if version is not None:
                    with withPrivatePipSitePackagesPathAdded(logger=logger):
                        ok, _message = _checkRequiredVersion(
                            logger=logger, tool=binary_name, tool_call=[possible]
                        )
                    if not ok:
                        continue

                return possible, site_packages_folder, assume_yes_for_downloads

    return None, site_packages_folder, assume_yes_for_downloads


def getPrivatePipBaseFolder():
    """Get the inline copy folder for a given name."""
    return os.path.join(
        getDownloadCacheDir(),
        "pip",
        "private-%s" % _getPrivatePipHash(),
    )


_private_pip_site_packages_dir = None


def getPrivatePipSitePackagesDir(logger):
    """Get the site-packages directory of the private pip space."""
    # We use the global statement to cache the result across calls.
    # pylint: disable=global-statement
    global _private_pip_site_packages_dir

    if _private_pip_site_packages_dir is None:
        download_folder = getPrivatePipBaseFolder()

        for root, dirnames, _filenames in os.walk(download_folder):
            found_candidate = None

            for candidate in getSitePackageCandidateNames():
                if candidate in dirnames:
                    # Unclear which one to use.
                    if found_candidate is not None:
                        return logger.sysexit(
                            "Scan for pip folder found multiple candidates: %s and %s."
                            % (found_candidate, candidate)
                        )

                    found_candidate = candidate

            if found_candidate:
                _private_pip_site_packages_dir = os.path.join(root, found_candidate)
                break

    return _private_pip_site_packages_dir


def withPrivatePipSitePackagesPathAdded(logger):
    """Context manager to add private pip site-packages to PYTHONPATH."""
    return withEnvironmentPathAdded(
        "PYTHONPATH", getPrivatePipSitePackagesDir(logger=logger), prefix=True
    )


def _isPackageInstalled(site_packages_folder, package_name, package_version):
    """Check if a package is installed in the given site-packages folder."""
    if package_version is None:
        return True

    # Basic check for dist-info specific to version
    dist_info_name = "%s-%s" % (
        package_name.replace("-", "_"),
        package_version,
    )

    dist_info_path = getNormalizedPathJoin(
        site_packages_folder,
        "%s.dist-info" % dist_info_name,
    )
    egg_info_path = getNormalizedPathJoin(
        site_packages_folder,
        "%s.egg-info" % dist_info_name,
    )

    if not os.path.isdir(dist_info_path) and not os.path.isdir(egg_info_path):
        return False
    else:
        # Metadata found, but check for conflicts with other versions,
        # e.g. when an upgrade was done, but not fully clean.
        for filename in os.listdir(site_packages_folder):
            if (
                filename.startswith(package_name + "-")
                and (filename.endswith(".dist-info") or filename.endswith(".egg-info"))
                and filename != os.path.basename(dist_info_path)
                and filename != os.path.basename(egg_info_path)
            ):
                return False

    return True


def tryDownloadPackageName(
    logger,
    package_name,
    module_name,
    package_version,
    assume_yes_for_downloads,
    reject_message,
):
    """Try to download a package from the private pip space."""
    download_folder = getPrivatePipBaseFolder()

    site_packages_folder = getPrivatePipSitePackagesDir(logger=logger)

    if site_packages_folder is not None:
        candidate = os.path.join(site_packages_folder, module_name)

        if os.path.exists(candidate):
            # If version is specified, check if it looks installed.
            if _isPackageInstalled(
                site_packages_folder=site_packages_folder,
                package_name=package_name,
                package_version=package_version,
            ):
                return (
                    site_packages_folder,
                    assume_yes_for_downloads,
                )

    if not _checkPackageConstraint(logger, package_name):
        return None, assume_yes_for_downloads

    if shouldDownload(
        message="Nuitka depends on '%s' to compile code, it is recommended."
        % package_name,
        reject_message=reject_message,
        assume_yes_for_downloads=assume_yes_for_downloads,
        download_ok=True,
    ):
        assume_yes_for_downloads = True

        if package_version is not None:
            package_spec = "%s==%s" % (package_name, package_version)
        else:
            package_spec = package_name

        command = (
            getSystemPrefixExecutable(),
            "-m",
            "pip",
            "install",
            "--no-warn-script-location",
            "--disable-pip-version-check",
            "--ignore-installed",
            "--upgrade",
            "--root",
            download_folder,
            "--prefix",
            ".",
            package_spec,
        )

        interactive = sys.stdout.isatty()

        with getNullOutput() as null_output:
            exit_code = subprocess.call(
                command,
                shell=False,
                stdout=None if interactive else null_output,
                stderr=None if interactive else null_output,
            )

        if exit_code != 0:
            return None, assume_yes_for_downloads

        if site_packages_folder is None:
            site_packages_folder = getPrivatePipSitePackagesDir(logger=logger)

        if site_packages_folder is None or not _isPackageInstalled(
            site_packages_folder=site_packages_folder,
            package_name=package_name,
            package_version=package_version,
        ):
            return logger.sysexit(
                "Error, failed to download package %r into private pip space."
                % package_name
            )

    if site_packages_folder is not None:
        return (
            site_packages_folder,
            assume_yes_for_downloads,
        )

    return None, assume_yes_for_downloads


def _evaluateConstraint(logger, constraint, package_name):
    """Evaluate a package constraint."""
    try:
        # We trust the file content, pylint: disable=eval-used
        return eval(
            constraint,
            {
                "python_version": _PythonVersion(python_version_str),
                "os": os,
                "sys": sys,
            },
        )
    except Exception as e:  # pylint: disable=broad-except
        return logger.sysexit(
            "Error, checking constraint %r for package %r gave error: %r"
            % (constraint, package_name, e)
        )


def _checkPackageConstraint(logger, package_name):
    """Check if a package constraint is met."""
    # Make sure we have the constraints loaded.
    _parseRequirements(logger)

    constraints = _private_pip_constraints.get(package_name)

    if not constraints:
        return True

    for constraint in constraints:
        if _evaluateConstraint(logger, constraint, package_name):
            return True

    return False


def getPrivatePackage(
    logger,
    package_name,
    module_name,
    package_version,
    submodule_names,
    assume_yes_for_downloads,
    reject_message,
):
    """Get a package from the private pip space or globally.

    Args:
        package_name: name of the package to download
        module_name: name of the module to import
        package_version: version of the package to download
        submodule_names: tuple of submodule names to import before returning
        assume_yes_for_downloads: if tools should be downloaded automatically

    Returns:
        module: the imported module or None if not found
    """
    try:
        module = __import__(module_name, fromlist=["*"])

        if submodule_names:
            for submodule_name in submodule_names:
                __import__(module_name + "." + submodule_name)

        return module
    except ImportError:
        site_packages_folder, _assume_yes_for_downloads = tryDownloadPackageName(
            logger=logger,
            package_name=package_name,
            module_name=module_name.split(".")[0],
            package_version=package_version,
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message=reject_message,
        )

        if site_packages_folder:
            with withTemporarySysPathExtension([site_packages_folder]):
                module = __import__(module_name, fromlist=["*"])

                if submodule_names:
                    for submodule_name in submodule_names:
                        __import__(module_name + "." + submodule_name)

            return module

        return None


def getZigBinaryPath(logger, assume_yes_for_downloads, reject_message):
    """Get the path of the Zig binary from the private pip space."""
    # spell-checker: ignore ziglang

    site_packages_folder, _assume_yes_for_downloads = tryDownloadPackageName(
        logger=logger,
        package_name="ziglang",
        module_name="ziglang",
        package_version=None,
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message=reject_message,
    )

    # The "ziglang" package puts a "python-zig.exe" in the scripts folder, and
    # not found due to mismatch, but the actual "zig.exe" is in the site-packages
    # folder.
    if site_packages_folder:
        zig_exe_path = getNormalizedPathJoin(site_packages_folder, "ziglang", "zig")

        if isWin32Windows():
            zig_exe_path += ".exe"

        if os.path.exists(zig_exe_path):
            return zig_exe_path

    return None


def getClangFormatBinaryPath(logger, assume_yes_for_downloads, reject_message):
    """Get the path of the clang-format binary from the private pip space."""
    binary_path, _site_packages_folder, _assume_yes_for_downloads = (
        _getPrivatePipBinaryPath(
            logger=logger,
            binary_name="clang-format",
            package_name="clang-format",
            module_name="clang_format",
            version=_getRequiredVersion(logger, "clang-format"),
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message=reject_message,
        )
    )
    return binary_path


def getBlackBinaryPath(logger, assume_yes_for_downloads):
    """Get the path of the black binary from the private pip space."""
    binary_path, _site_packages_folder, _assume_yes_for_downloads = (
        _getPrivatePipBinaryPath(
            logger=logger,
            binary_name="black",
            package_name="black",
            module_name="black",
            version=_getRequiredVersion(logger, "black"),
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="Python formatting needs to use 'black'.",
        )
    )
    return binary_path


def getIsortBinaryPath(logger, assume_yes_for_downloads):
    """Get the path of the isort binary from the private pip space."""
    binary_path, _site_packages_folder, _assume_yes_for_downloads = (
        _getPrivatePipBinaryPath(
            logger=logger,
            binary_name="isort",
            package_name="isort",
            module_name="isort",
            version=_getRequiredVersion(logger, "isort"),
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="Python formatting needs to use 'isort'.",
        )
    )
    return binary_path


def getMdformatBinaryPath(logger, assume_yes_for_downloads):
    """Get the path of the mdformat binary from the private pip space."""
    mdformat_path, site_packages_folder, assume_yes_for_downloads = (
        _getPrivatePipBinaryPath(
            logger=logger,
            binary_name="mdformat",
            package_name="mdformat",
            module_name="mdformat",
            version=_getRequiredVersion(logger, "mdformat"),
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="Markdown formatting needs to use 'mdformat'.",
        )
    )

    if mdformat_path and site_packages_folder:
        tryDownloadPackageName(
            logger=logger,
            package_name="mdformat-gfm",
            module_name="mdformat_gfm",
            package_version=_getRequiredVersion(logger, "mdformat-gfm"),
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="Markdown formatting needs to use 'mdformat-gfm'.",
        )
        tryDownloadPackageName(
            logger=logger,
            package_name="mdformat-frontmatter",
            module_name="mdformat_frontmatter",
            package_version=_getRequiredVersion(logger, "mdformat-frontmatter"),
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="Markdown formatting needs to use 'mdformat-frontmatter'.",
        )
        tryDownloadPackageName(
            logger=logger,
            package_name="mdformat-footnote",
            module_name="mdformat_footnote",
            package_version=_getRequiredVersion(logger, "mdformat-footnote"),
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="Markdown formatting needs to use 'mdformat-footnote'.",
        )

    return mdformat_path


def getRstfmtBinaryPath(logger, assume_yes_for_downloads):
    """Get the path of the rstfmt binary from the private pip space."""
    binary_path, _site_packages_folder, _assume_yes_for_downloads = (
        _getPrivatePipBinaryPath(
            logger=logger,
            binary_name="rstfmt",
            package_name="rstfmt",
            module_name="rstfmt",
            version=_getRequiredVersion(logger, "rstfmt"),
            assume_yes_for_downloads=assume_yes_for_downloads,
            reject_message="ReStructuredText formatting needs to use 'rstfmt'.",
        )
    )
    return binary_path


@total_ordering
class _PythonVersion(object):
    def __init__(self, version_str):
        self.version = tuple(int(x) for x in version_str.split("."))

    @staticmethod
    def _makeTuple(other):
        if str is not bytes and isinstance(other, (str, bytes)):
            return tuple(int(x) for x in other.split("."))
        return other

    def __eq__(self, other):
        return self.version == self._makeTuple(other)

    def __lt__(self, other):
        return self.version < self._makeTuple(other)


_private_pip_requirements = None
_private_pip_constraints = None


def _parseRequirements(logger):
    """Parse the requirements-private.txt file."""
    # We use the global statement to cache the result across calls.
    # pylint: disable=global-statement
    global _private_pip_requirements, _private_pip_constraints

    if _private_pip_requirements is None:
        _private_pip_requirements = OrderedDict()
        _private_pip_constraints = OrderedDict()

        for line in getFileContentByLine(
            os.path.join(os.path.dirname(__file__), "requirements-private.txt")
        ):
            line = line.strip()

            if not line or line.startswith("#"):
                continue

            parts = re.split(r"\s*;\s*", line, maxsplit=1)

            if len(parts) > 1:
                tool_part = parts[0]
                constraint_part = parts[1]
            else:
                tool_part = line
                constraint_part = None

            if "==" in tool_part:
                tool_name = tool_part.split("==")[0].strip()
            else:
                for other_comparison in (">=", "<=", ">", "<", "~="):
                    if other_comparison in tool_part:
                        msg = (
                            """\
Error, invalid requirement format in 'requirements-private.txt' line %r, only '==' is allowed."""
                            % line
                        )

                        return logger.sysexit(msg)

                tool_name = tool_part.strip()

            if tool_name not in _private_pip_requirements:
                _private_pip_requirements[tool_name] = []
                _private_pip_constraints[tool_name] = []

            _private_pip_requirements[tool_name].append(line)
            if constraint_part is not None:
                _private_pip_constraints[tool_name].append(constraint_part)

    return _private_pip_requirements


def _getPrivatePipHash():
    """Calculate the hash of the private pip space."""
    private_pip_hash = HashCRC32()
    private_pip_hash.updateFromValues(
        python_version_str,
        getOS(),
        getArchitecture(),
    )
    private_pip_hash.updateFromValues(_parseRequirements(logger=None))

    return private_pip_hash.asHexDigest()


def _getRequiredVersion(logger, tool):
    """Get the required version of a tool from requirements-private.txt."""
    requirements = _parseRequirements(logger)

    if tool not in requirements:
        return logger.sysexit(
            "Error, cannot find %r in 'requirements-private.txt' file." % tool
        )

    for line in requirements[tool]:
        parts = re.split(r"\s*;\s*", line, maxsplit=1)

        if len(parts) > 1:
            constraint = parts[1]
            if not _evaluateConstraint(logger, constraint, tool):
                continue

        tool_part = parts[0]

        # TODO: This is not quite correct for all packages, but we only have equality
        # checks right now.
        if "==" in tool_part:
            return tool_part.split("==")[1].strip()
        else:
            return None

    return None


def _checkRequiredVersion(logger, tool, tool_call):
    """Check if the version of a tool matches the requirement."""
    required_version = _getRequiredVersion(logger=logger, tool=tool)

    tool_call = list(tool_call) + ["--version"]

    try:
        version_output = check_output(tool_call)
    except NuitkaCalledProcessError as e:
        return False, "failed to execute: %s" % e.stderr

    if str is not bytes:
        version_output = version_output.decode("utf8")

    actual_version = None
    for line in version_output.splitlines():
        line = line.strip()

        if line.startswith(
            ("black, ", "python -m black,", "__main__.py, ", "mdformat ")
        ):
            if "(" in line:
                line = line[: line.rfind("(")].strip()

            actual_version = line.split()[-1]
            break
        if line.startswith("VERSION "):
            actual_version = line.split()[-1]
            break
        # spell-checker: ignore rstfmt
        if line.startswith("rstfmt "):
            actual_version = line.split()[-1]
            break
        if line.startswith("clang-format version "):
            actual_version = line.split()[2]
            break

    if not actual_version:
        return False, "Error, couldn't determine version output of '%s' ('%s')" % (
            tool,
            " ".join(tool_call),
        )

    message = "Version of '%s' via '%s' is required to be %r and not %r." % (
        tool,
        " ".join(tool_call),
        required_version,
        actual_version,
    )

    return required_version == actual_version, message


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Generate header files that provide compiler specifics."""

from nuitka.build.SconsInterface import (
    cleanSconsDirectory,
    getCommonSconsOptions,
    provideStaticSourceFilesOffsets,
    runScons,
)
from nuitka.PythonVersions import isPythonWithGil, python_version_str
from nuitka.Tracing import offsets_logger
from nuitka.utils.Execution import check_output
from nuitka.utils.FileOperations import makePath, withTemporaryFilename


def generateHeader():
    scons_options, env_values = getCommonSconsOptions()

    scons_options["source_dir"] = "generate_header.build"
    cleanSconsDirectory(scons_options["source_dir"])
    makePath(scons_options["source_dir"])

    provideStaticSourceFilesOffsets(scons_options["source_dir"])

    python_version_id = "%s_%s" % (
        python_version_str,
        "gil" if isPythonWithGil() else "no-gil",
    )

    with withTemporaryFilename(prefix=python_version_id, suffix=".exe") as result_exe:
        scons_options["result_exe"] = result_exe

        runScons(
            scons_options=scons_options,
            env_values=env_values,
            scons_filename="Offsets.scons",
        )

        header_output = check_output([result_exe])

        if str is not bytes:
            header_output = header_output.decode("utf8")

        offsets_logger.info(repr(header_output))

        lines = header_output.splitlines()

        if lines[-1] != "OK.":
            offsets_logger.sysexit("Error, failed to produce expected output.")
        del lines[-1]

        for line in lines:
            offsets_logger.info("Processing: %s" % line)

        offsets_logger.info("OK.")


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com

"""Nuitka version related stuff."""

version_string = """\
Nuitka V4.0.2
Copyright (C) 2025 Kay Hayen."""


def getNuitkaVersion():
    """Return Nuitka version as a string.

    This should not be used for >= comparisons directly.
    """
    return version_string.split()[1][1:]


# Sanity check.
assert getNuitkaVersion()[-1].isdigit(), getNuitkaVersion()


def parseNuitkaVersionToTuple(version):
    """Return Nuitka version as a tuple.

    This can also not be used for precise comparisons, even with rc versions,
    but it's not actually a version.
    """

    if "rc" in version:
        rc_number = int(version[version.find("rc") + 2 :] or "0")
        version = version[: version.find("rc")]

        is_final = False
    else:
        rc_number = 0
        is_final = True

    result = version.split(".")
    if len(result) == 2:
        result.append("0")

    result = [int(digit) for digit in result]
    result.extend((is_final, rc_number))
    return tuple(result)


def getNuitkaVersionTuple():
    """Return Nuitka version as a tuple.

    This can also not be used for precise comparisons, even with rc versions,
    but it's not actually a version. The format is used what is used for
    "__compiled__" values.
    """

    return parseNuitkaVersionToTuple(version=getNuitkaVersion())


def getNuitkaVersionYear():
    """The year of Nuitka copyright for use in generations."""

    return int(version_string.split()[4])


def getCommercialVersion():
    """Return Nuitka commercial version if installed."""
    try:
        from nuitka.tools.commercial import Version
    except ImportError:
        return None
    else:
        return Version.__version__


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Nodes for statements."""

from nuitka.PythonVersions import python_version

from .NodeBases import StatementBase
from .StatementBasesGenerated import (
    StatementExpressionOnlyBase,
    StatementsSequenceBase,
)


class StatementsSequenceMixin(object):
    __slots__ = ()

    def finalize(self):
        del self.parent

        for s in self.subnode_statements:
            s.finalize()

    @staticmethod
    def isStatementsSequence():
        return True

    def computeStatement(self, trace_collection):
        # Don't want to be called like this.
        assert False, self

    def trimStatements(self, statement):
        assert statement.parent is self

        old_statements = list(self.subnode_statements)
        assert statement in old_statements, (statement, self)

        new_statements = old_statements[: old_statements.index(statement) + 1]

        self.setChildStatements(new_statements)

    def removeStatement(self, statement):
        assert statement.parent is self

        statements = list(self.subnode_statements)
        statements.remove(statement)
        self.setChildStatements(tuple(statements))

        if statements:
            return self
        else:
            return None

    def replaceStatement(self, statement, statements):
        old_statements = list(self.subnode_statements)

        merge_index = old_statements.index(statement)

        new_statements = (
            tuple(old_statements[:merge_index])
            + tuple(statements)
            + tuple(old_statements[merge_index + 1 :])
        )

        self.setChildStatements(new_statements)

    def mayHaveSideEffects(self):
        # Statement sequences have a side effect if one of the statements does.
        for statement in self.subnode_statements:
            if statement.mayHaveSideEffects():
                return True
        return False

    def mayRaiseException(self, exception_type):
        for statement in self.subnode_statements:
            if statement.mayRaiseException(exception_type):
                return True
        return False

    def needsFrame(self):
        for statement in self.subnode_statements:
            if statement.needsFrame():
                return True
        return False

    def mayReturn(self):
        for statement in self.subnode_statements:
            if statement.mayReturn():
                return True
        return False

    def mayBreak(self):
        for statement in self.subnode_statements:
            if statement.mayBreak():
                return True
        return False

    def mayContinue(self):
        for statement in self.subnode_statements:
            if statement.mayContinue():
                return True
        return False

    def mayRaiseExceptionOrAbort(self, exception_type):
        return (
            self.mayRaiseException(exception_type)
            or self.mayReturn()
            or self.mayBreak()
            or self.mayContinue()
        )

    def isStatementAborting(self):
        return self.subnode_statements[-1].isStatementAborting()

    def willRaiseAnyException(self):
        return self.subnode_statements[-1].willRaiseAnyException()


class StatementsSequence(StatementsSequenceMixin, StatementsSequenceBase):
    kind = "STATEMENTS_SEQUENCE"

    named_children = ("statements|tuple+setter",)

    @staticmethod
    def isStatementsSequenceButNotFrame():
        return True

    def computeStatementsSequence(self, trace_collection):
        new_statements = []

        statements = self.subnode_statements
        assert statements, self

        for count, statement in enumerate(statements):
            # May be frames embedded.
            if statement.isStatementsFrame():
                new_statement = statement.computeStatementsSequence(trace_collection)
            else:
                new_statement = trace_collection.onStatement(statement=statement)

            if new_statement is not None:
                if new_statement.isStatementsSequenceButNotFrame():
                    new_statements.extend(new_statement.subnode_statements)
                else:
                    new_statements.append(new_statement)

                if (
                    statement is not statements[-1]
                    and new_statement.isStatementAborting()
                ):
                    trace_collection.signalChange(
                        "new_statements",
                        statements[count + 1].getSourceReference(),
                        "Removed dead statements.",
                    )

                    for s in statements[statements.index(statement) + 1 :]:
                        s.finalize()

                    break

        new_statements = tuple(new_statements)
        if statements != new_statements:
            if new_statements:
                self.setChildStatements(new_statements)

                return self
            else:
                return None
        else:
            return self

    @staticmethod
    def getStatementNiceName():
        return "statements sequence"


class StatementExpressionOnly(StatementExpressionOnlyBase):
    kind = "STATEMENT_EXPRESSION_ONLY"

    named_children = ("expression",)

    def mayHaveSideEffects(self):
        return self.subnode_expression.mayHaveSideEffects()

    def mayRaiseException(self, exception_type):
        return self.subnode_expression.mayRaiseException(exception_type)

    def computeStatement(self, trace_collection):
        # TODO: Maybe also have a variant that will not attempt dropping anymore, for nodes
        # that are known to not do it anymore, or wait for a node change, or make it part of
        # the expression interface to be used in statement form as well.

        expression = trace_collection.onExpression(self.subnode_expression)

        return expression.computeExpressionDrop(
            statement=self, trace_collection=trace_collection
        )

    @staticmethod
    def getStatementNiceName():
        return "expression only statement"

    def getDetailsForDisplay(self):
        return {"expression": self.subnode_expression.kind}


class StatementPreserveFrameException(StatementBase):
    kind = "STATEMENT_PRESERVE_FRAME_EXCEPTION"

    __slots__ = ("preserver_id",)

    def __init__(self, preserver_id, source_ref):
        StatementBase.__init__(self, source_ref=source_ref)

        self.preserver_id = preserver_id

    def finalize(self):
        del self.parent

    def getDetails(self):
        return {"preserver_id": self.preserver_id}

    def getPreserverId(self):
        return self.preserver_id

    if python_version < 0x300:

        def computeStatement(self, trace_collection):
            # For Python2 generators, it's not necessary to preserve, the frame
            # decides it. TODO: This check makes only sense once.

            if self.getParentStatementsFrame().needsExceptionFramePreservation():
                return self, None, None
            else:
                return (
                    None,
                    "new_statements",
                    "Removed frame preservation for generators.",
                )

    else:

        def computeStatement(self, trace_collection):
            return self, None, None

    @staticmethod
    def mayRaiseException(exception_type):
        return False

    @staticmethod
    def needsFrame():
        return True


class StatementRestoreFrameException(StatementBase):
    kind = "STATEMENT_RESTORE_FRAME_EXCEPTION"

    __slots__ = ("preserver_id",)

    def __init__(self, preserver_id, source_ref):
        StatementBase.__init__(self, source_ref=source_ref)

        self.preserver_id = preserver_id

    def finalize(self):
        del self.parent

    def getDetails(self):
        return {"preserver_id": self.preserver_id}

    def getPreserverId(self):
        return self.preserver_id

    def computeStatement(self, trace_collection):
        return self, None, None

    @staticmethod
    def mayRaiseException(exception_type):
        return False


class StatementPublishException(StatementBase):
    kind = "STATEMENT_PUBLISH_EXCEPTION"

    def __init__(self, source_ref):
        StatementBase.__init__(self, source_ref=source_ref)

    def finalize(self):
        del self.parent

    def computeStatement(self, trace_collection):
        # TODO: Determine the need for it.
        return self, None, None

    @staticmethod
    def mayRaiseException(exception_type):
        return False


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Assignment related nodes.

The most simple assignment statement ``a = b`` is what we have here. All others
are either re-formulated using temporary variables, e.g. ``a, b = c`` or are
attribute, slice, subscript assignments.

The deletion is a separate node unlike in CPython where assigning to ``NULL`` is
internally what deletion is. But deleting is something entirely different to us
during code generation, which is why we keep them separate.

Tracing assignments in SSA form is the core of optimization for which we use
the traces.

"""

from abc import abstractmethod

from nuitka.ModuleRegistry import getOwnerFromCodeName
from nuitka.options.Options import isExperimental

from .ConstantRefNodes import makeConstantRefNode
from .NodeMakingHelpers import (
    makeStatementExpressionOnlyReplacementNode,
    makeStatementsSequenceReplacementNode,
)
from .shapes.ControlFlowDescriptions import (
    ControlFlowDescriptionElementBasedEscape,
    ControlFlowDescriptionFullEscape,
    ControlFlowDescriptionNoEscape,
)
from .shapes.IteratorShapes import tshape_iterator
from .shapes.StandardShapes import tshape_unknown
from .StatementBasesGenerated import (
    StatementAssignmentVariableConstantImmutableBase,
    StatementAssignmentVariableConstantMutableBase,
    StatementAssignmentVariableFromTempVariableBase,
    StatementAssignmentVariableFromVariableBase,
    StatementAssignmentVariableGenericBase,
    StatementAssignmentVariableHardValueBase,
    StatementAssignmentVariableIteratorBase,
)
from .VariableDelNodes import makeStatementDelVariable
from .VariableRefNodes import ExpressionTempVariableRef


class StatementAssignmentVariableMixin(object):
    """Assignment to a variable from an expression.

    All assignment forms that are not to attributes, slices, subscripts
    use this.

    The source might be a complex expression. The target can be any kind
    of variable, temporary, local, global, etc.

    Assigning a variable is something we trace in a new version, this is
    hidden behind target variable reference, which has this version once
    it can be determined.
    """

    # Mixins are not allowed to specify slots, pylint: disable=assigning-non-slot
    __slots__ = ()

    @staticmethod
    def isStatementAssignmentVariable():
        return True

    def finalize(self):
        del self.variable
        del self.variable_trace
        self.subnode_source.finalize()
        del self.subnode_source

    def getDetailsForDisplay(self):
        return {
            "variable_name": self.getVariableName(),
            "is_temp": self.variable.isTempVariable(),
            "var_type": self.variable.getVariableType(),
            "owner": self.variable.getOwner().getCodeName(),
        }

    @classmethod
    def fromXML(cls, provider, source_ref, **args):
        # Virtual method overload, pylint: disable=unused-argument
        owner = getOwnerFromCodeName(args["owner"])

        # TODO: Need to also provide "outline" for best effect
        if args["is_temp"] == "True":
            variable = owner.createTempVariable(
                args["variable_name"], temp_type=["var_type"], outline=None
            )
        else:
            variable = owner.getProvidedVariable(args["variable_name"])

        del args["is_temp"]
        del args["var_type"]
        del args["owner"]

        version = variable.allocateTargetNumber()

        return cls(variable=variable, version=version, source_ref=source_ref, **args)

    def makeClone(self):
        version = self.variable.allocateTargetNumber()

        return self.__class__(
            source=self.subnode_source.makeClone(),
            variable=self.variable,
            variable_version=version,
            source_ref=self.source_ref,
        )

    def getVariableName(self):
        return self.variable.getName()

    def getVariable(self):
        return self.variable

    def setVariable(self, variable):
        self.variable = variable
        self.variable_version = variable.allocateTargetNumber()

    def getVariableTrace(self):
        return self.variable_trace

    def markAsInplaceSuspect(self):
        self.inplace_suspect = True

    def isInplaceSuspect(self):
        return self.inplace_suspect

    def removeMarkAsInplaceSuspect(self):
        self.inplace_suspect = False

    def mayRaiseException(self, exception_type):
        return self.subnode_source.mayRaiseException(exception_type)

    def needsReleasePreviousValue(self):
        previous = self.variable_trace.getPrevious()

        if previous.mustNotHaveValue():
            return False
        elif previous.mustHaveValue():
            return True
        else:
            return None

    @staticmethod
    def getStatementNiceName():
        return "variable assignment statement"

    def getTypeShape(self):
        # Might be finalized, e.g. due to being dead code.
        try:
            source = self.subnode_source
        except AttributeError:
            return tshape_unknown

        return source.getTypeShape()

    @staticmethod
    def mayHaveSideEffects():
        # TODO: May execute "__del__" code, it would be sweet to be able to predict
        # if another reference will still be active for a value though, or if there
        # is such a code for the type shape.
        return True

    def _considerSpecialization(self, old_source, source):
        if source.isCompileTimeConstant():
            result = makeStatementAssignmentVariableConstant(
                source=source,
                variable=self.variable,
                variable_version=self.variable_version,
                very_trusted=old_source.isExpressionImportName(),
                source_ref=self.source_ref,
            )

            return (
                result,
                "new_statements",
                "Assignment source of '%s' is now compile time constant."
                % self.getVariableName(),
            )

        if source.isExpressionVariableRef():
            result = StatementAssignmentVariableFromVariable(
                source=source,
                variable=self.variable,
                variable_version=self.variable_version,
                source_ref=self.source_ref,
            )

            return (
                result,
                "new_statements",
                "Assignment source is now variable reference.",
            )

        if source.isExpressionTempVariableRef():
            result = StatementAssignmentVariableFromTempVariable(
                source=source,
                variable=self.variable,
                variable_version=self.variable_version,
                source_ref=self.source_ref,
            )

            return (
                result,
                "new_statements",
                "Assignment source is now temp variable reference.",
            )

        if source.getTypeShape().isShapeIterator():
            result = StatementAssignmentVariableIterator(
                source=source,
                variable=self.variable,
                variable_version=self.variable_version,
                source_ref=self.source_ref,
            )

            return (
                result,
                "new_statements",
                "Assignment source is now known to be iterator.",
            )

        if source.hasVeryTrustedValue():
            result = StatementAssignmentVariableHardValue(
                source=source,
                variable=self.variable,
                variable_version=self.variable_version,
                source_ref=self.source_ref,
            )

            return (
                result,
                "new_statements",
                "Assignment source is now known to be hard value.",
            )

        return self, None, None

    def collectVariableAccesses(self, emit_variable):
        emit_variable(self.variable)
        self.subnode_source.collectVariableAccesses(emit_variable)

    @abstractmethod
    def hasVeryTrustedValue(self):
        """Does this assignment node have a very trusted value."""

    @abstractmethod
    def computeStatementAssignmentTraceUpdate(self, trace_collection, source):
        """Set the value trace for the assignment form."""


class StatementAssignmentVariableGeneric(
    StatementAssignmentVariableMixin, StatementAssignmentVariableGenericBase
):
    kind = "STATEMENT_ASSIGNMENT_VARIABLE_GENERIC"

    named_children = ("source|setter",)
    nice_children = ("assignment source",)
    node_attributes = ("variable", "variable_version")
    auto_compute_handling = "post_init"

    __slots__ = (
        "variable_trace",
        # TODO: Does every variant have to care here
        "inplace_suspect",
    )

    # False alarm due to post_init, pylint: disable=attribute-defined-outside-init

    # TODO: Add parsing of node_attributes for init values, so we can avoid these too
    def postInitNode(self):
        self.variable_trace = None
        self.inplace_suspect = None

    @staticmethod
    def getReleaseEscape():
        return ControlFlowDescriptionFullEscape

    def computeStatement(self, trace_collection):
        # TODO: Way too ugly to have global trace kinds just here, and needs to
        # be abstracted somehow. But for now we let it live here.
        old_source = self.subnode_source
        variable = self.variable

        if old_source.isExpressionSideEffects():
            # If the assignment source has side effects, we can put them into a
            # sequence and compute that instead.
            statements = [
                makeStatementExpressionOnlyReplacementNode(side_effect, self)
                for side_effect in old_source.subnode_side_effects
            ]

            statements.append(self)

            # Remember out parent, we will assign it for the sequence to use.
            parent = self.parent

            # Need to update ourselves to no longer reference the side effects,
            # but go to the wrapped thing.
            self.setChildSource(old_source.subnode_expression)

            result = makeStatementsSequenceReplacementNode(
                statements=statements, node=self
            )
            result.parent = parent

            return (
                result.computeStatementsSequence(trace_collection),
                "new_statements",
                """\
Side effects of assignments promoted to statements.""",
            )

        # Let assignment source may re-compute first.
        source = trace_collection.onExpression(self.subnode_source)

        # No assignment will occur, if the assignment source raises, so give up
        # on this, and return it as the only side effect.
        if source.willRaiseAnyException():
            result = makeStatementExpressionOnlyReplacementNode(
                expression=source, node=self
            )

            del self.parent

            return (
                result,
                "new_raise",
                """\
Assignment raises exception in assigned value, removed assignment.""",
            )

        # Assigning from and to the same variable, can be optimized away
        # immediately, there is no point in doing it. Exceptions are of course
        # module variables that collide with built-in names.

        # TODO: The variable type checks ought to become unnecessary, as they
        # are to be a feature of the trace. Assigning from known assigned is
        # supposed to be possible to eliminate. If we get that wrong, we are
        # doing it wrong.
        if (
            not variable.isModuleVariable()
            and source.isExpressionVariableRef()
            and source.getVariable() is variable
        ):
            # A variable access that has a side effect, must be preserved,
            # so it can e.g. raise an exception, otherwise we can be fully
            # removed.
            if source.mayHaveSideEffects():
                result = makeStatementExpressionOnlyReplacementNode(
                    expression=source, node=self
                )

                return (
                    result,
                    "new_statements",
                    """\
Lowered assignment of %s from itself to mere access of it."""
                    % variable.getDescription(),
                )
            else:
                return (
                    None,
                    "new_statements",
                    """\
Removed assignment of %s from itself which is known to be defined."""
                    % variable.getDescription(),
                )

        if source is old_source:
            result = self, None, None
        else:
            result = self._considerSpecialization(old_source, source)
            result[0].parent = self.parent

        result[0].computeStatementAssignmentTraceUpdate(trace_collection, source)

        return result

    def computeStatementAssignmentTraceUpdate(self, trace_collection, source):
        # Set-up the trace to the trace collection, so future references will
        # find this assignment.
        self.variable_trace = trace_collection.onVariableSet(
            variable=self.variable, version=self.variable_version, assign_node=self
        )

        # TODO: Determine from future use of assigned variable, if this is needed at all.
        trace_collection.removeKnowledge(source)

    def hasVeryTrustedValue(self):
        """Does this assignment node have a very trusted value."""
        return self.subnode_source.hasVeryTrustedValue()


class StatementAssignmentVariableIterator(
    StatementAssignmentVariableMixin, StatementAssignmentVariableIteratorBase
):
    # Carries state for propagating iterators potentially.

    # TODO: Maybe have a namedtuple with these intended for index replacement,
    # they form once set of things pylint: disable=too-many-instance-attributes

    kind = "STATEMENT_ASSIGNMENT_VARIABLE_ITERATOR"

    named_children = ("source",)
    nice_children = ("assignment source",)
    node_attributes = ("variable", "variable_version")
    auto_compute_handling = "post_init"

    __slots__ = (
        "variable_trace",
        # TODO: Does every variant have to care here
        "inplace_suspect",
        "type_shape",
        "temp_scope",
        "tmp_iterated_variable",
        "tmp_iteration_count_variable",
        "tmp_iteration_next_variable",
        "is_indexable",
    )

    # False alarm due to post_init, pylint: disable=attribute-defined-outside-init

    # TODO: Add parsing of node_attributes for init values, so we can avoid these too
    def postInitNode(self):
        self.variable_trace = None
        self.inplace_suspect = None

        # Have a valid start value, later this will become more specific.
        self.type_shape = tshape_iterator

        # For replacing with indexing potentially.
        self.temp_scope = None
        self.tmp_iterated_variable = None
        self.tmp_iteration_count_variable = None
        self.tmp_iteration_next_variable = None

        # When found non-indexable, we do not try again.
        self.is_indexable = None

    def getTypeShape(self):
        return self.type_shape

    @staticmethod
    def getReleaseEscape():
        # TODO: For iteration over constants that wouldn't be necessary,
        # but if we know the iteration well enough, it's supposed to be
        # converted to something else anyway.
        return ControlFlowDescriptionElementBasedEscape

    def getIterationIndexDesc(self):
        """For use in optimization outputs only, here and using nodes."""
        return "'%s[%s]'" % (
            self.tmp_iterated_variable.getName(),
            self.tmp_iteration_count_variable.getName(),
        )

    def _replaceWithDirectAccess(self, trace_collection, provider):
        self.temp_scope = provider.allocateTempScope("iterator_access")

        self.tmp_iterated_variable = provider.allocateTempVariable(
            temp_scope=self.temp_scope, name="iterated_value", temp_type="object"
        )

        reference_iterated = ExpressionTempVariableRef(
            variable=self.tmp_iterated_variable,
            source_ref=self.subnode_source.source_ref,
        )

        iterated_value = self.subnode_source.subnode_value

        assign_iterated = makeStatementAssignmentVariable(
            source=iterated_value,
            variable=self.tmp_iterated_variable,
            source_ref=iterated_value.source_ref,
        )

        self.tmp_iteration_count_variable = provider.allocateTempVariable(
            temp_scope=self.temp_scope, name="iteration_count", temp_type="object"
        )

        assign_iteration_count = makeStatementAssignmentVariable(
            source=makeConstantRefNode(constant=0, source_ref=self.source_ref),
            variable=self.tmp_iteration_count_variable,
            source_ref=iterated_value.source_ref,
        )

        # TODO: Unclear what node this really is right now, need to try out.
        self.subnode_source.setChildValue(reference_iterated)

        # Make sure variable trace is computed.
        assign_iterated.computeStatement(trace_collection)
        assign_iteration_count.computeStatement(trace_collection)
        reference_iterated.computeExpressionRaw(trace_collection)

        self.variable_trace = trace_collection.onVariableSet(
            variable=self.variable,
            version=self.variable_version,
            assign_node=self,
        )

        # For use when the "next" is replaced.
        self.tmp_iteration_next_variable = provider.allocateTempVariable(
            temp_scope=self.temp_scope, name="next_value", temp_type="object"
        )

        result = makeStatementsSequenceReplacementNode(
            (assign_iteration_count, assign_iterated, self), self
        )

        return (
            result,
            "new_statements",
            lambda: "Enabling indexing of iterated value through %s."
            % self.getIterationIndexDesc(),
        )

    def computeStatement(self, trace_collection):
        # TODO: Way too ugly to have global trace kinds just here, and needs to
        # be abstracted somehow. But for now we let it live here.
        source = self.subnode_source
        variable = self.variable

        provider = trace_collection.getOwner()

        # Let assignment source may re-compute first.
        source = trace_collection.onExpression(self.subnode_source)

        if (
            self.variable_trace is not None
            and self.tmp_iterated_variable is None
            and self.is_indexable is None
            and source.isExpressionBuiltinIterForUnpack()
            and isExperimental("iterator-optimization")
        ):
            if variable.hasAccessesOutsideOf(provider) is False:
                # TODO: Ought to become able to check if the self.variable_trace
                # is escaped by annotating that upwards, so a search is not
                # needed at all.
                last_trace = variable.getMatchingUnescapedAssignTrace(self)

                if last_trace is not None:
                    # Might not be allowed, remember if it's not allowed, otherwise retry.
                    self.is_indexable = (
                        source.subnode_value.getTypeShape().hasShapeIndexLookup()
                    )

                    if self.is_indexable:
                        return self._replaceWithDirectAccess(
                            trace_collection=trace_collection, provider=provider
                        )

        # No assignment will occur, if the assignment source raises, so give up
        # on this, and return it as the only side effect.
        if source.willRaiseAnyException():
            result = makeStatementExpressionOnlyReplacementNode(
                expression=source, node=self
            )

            del self.parent

            return (
                result,
                "new_raise",
                """\
Assignment raises exception in assigned value, removed assignment.""",
            )

        self.type_shape = source.getTypeShape()

        # Set-up the trace to the trace collection, so future references will
        # find this assignment.
        # Note: Keep this aligned with computeStatementAssignmentTraceUpdate
        self.variable_trace = trace_collection.onVariableSet(
            variable=variable, version=self.variable_version, assign_node=self
        )

        return self, None, None

    def computeStatementAssignmentTraceUpdate(self, trace_collection, source):
        self.variable_trace = trace_collection.onVariableSet(
            variable=self.variable, version=self.variable_version, assign_node=self
        )

    @staticmethod
    def hasVeryTrustedValue():
        """Does this assignment node have a very trusted value."""
        return False


class StatementAssignmentVariableConstantMutable(
    StatementAssignmentVariableMixin, StatementAssignmentVariableConstantMutableBase
):
    kind = "STATEMENT_ASSIGNMENT_VARIABLE_CONSTANT_MUTABLE"

    named_children = ("source",)
    nice_children = ("assignment source",)
    node_attributes = ("variable", "variable_version")
    auto_compute_handling = "post_init"

    __slots__ = (
        "variable_trace",
        # TODO: Does every variant have to care here
        "inplace_suspect",
    )

    # False alarm due to post_init, pylint: disable=attribute-defined-outside-init

    # TODO: Add parsing of node_attributes for init values, so we can avoid these too
    def postInitNode(self):
        self.variable_trace = None
        self.inplace_suspect = None

    @staticmethod
    def mayRaiseException(exception_type):
        return False

    @staticmethod
    def getReleaseEscape():
        return ControlFlowDescriptionNoEscape

    def computeStatement(self, trace_collection):
        variable = self.variable
        last_trace = self.variable_trace

        # Set-up the trace to the trace collection, so future references will
        # find this assignment.
        # Note: Keep this aligned with computeStatementAssignmentTraceUpdate

        provider = trace_collection.getOwner()

        if (
            last_trace is not None
            and last_trace.hasNoMergeOrNameUsage()
            and variable.hasAccessesOutsideOf(provider) is False
        ):
            if (
                variable.isModuleVariable()
                or variable.owner.locals_scope.isUnoptimizedFunctionScope()
            ):
                # TODO: We do not trust these yet a lot, but more might be
                pass
            else:
                # Unused constants can be eliminated in any case.
                if not last_trace.getUsageCount():
                    if last_trace.getPrevious().isUnassignedTrace():
                        return (
                            None,
                            "new_statements",
                            "Dropped dead assignment statement to '%s'."
                            % (self.getVariableName()),
                        )
                    else:
                        return trace_collection.computedStatementResult(
                            statement=makeStatementDelVariable(
                                variable=self.variable,
                                version=self.variable_version,
                                tolerant=True,
                                source_ref=self.source_ref,
                            ),
                            change_tags="new_statements",
                            change_desc="Lowered dead assignment statement to '%s' to previous value 'del'."
                            % self.getVariableName(),
                        )

        # Can safely forward propagate only non-mutable constants.
        self.variable_trace = trace_collection.onVariableSet(
            variable=variable, version=self.variable_version, assign_node=self
        )

        return self, None, None

    def computeStatementAssignmentTraceUpdate(self, trace_collection, source):
        self.variable_trace = trace_collection.onVariableSet(
            variable=self.variable, version=self.variable_version, assign_node=self
        )

    @staticmethod
    def hasVeryTrustedValue():
        """Does this assignment node have a very trusted value."""
        return False


class StatementAssignmentVariableConstantImmutable(
    StatementAssignmentVariableMixin, StatementAssignmentVariableConstantImmutableBase
):
    kind = "STATEMENT_ASSIGNMENT_VARIABLE_CONSTANT_IMMUTABLE"

    named_children = ("source",)
    nice_children = ("assignment source",)
    node_attributes = ("variable", "variable_version")
    auto_compute_handling = "post_init"

    __slots__ = (
        "variable_trace",
        # TODO: Does every variant have to care here
        "inplace_suspect",
    )

    # False alarm due to post_init, pylint: disable=attribute-defined-outside-init

    # TODO: Add parsing of node_attributes for init values, so we can avoid these too
    def postInitNode(self):
        self.variable_trace = None
        self.inplace_suspect = None

    @staticmethod
    def mayRaiseException(exception_type):
        return False

    @staticmethod
    def getReleaseEscape():
        return ControlFlowDescriptionNoEscape

    def computeStatement(self, trace_collection):
        variable = self.variable
        last_trace = self.variable_trace

        # Set-up the trace to the trace collection, so future references will
        # find this assignment.
        provider = trace_collection.getOwner()

        if (
            last_trace is not None
            and last_trace.hasNoMergeOrNameUsage()
            and variable.hasAccessesOutsideOf(provider) is False
        ):
            if (
                variable.isModuleVariable()
                or variable.owner.locals_scope.isUnoptimizedFunctionScope()
            ):
                # TODO: We do not trust these yet a lot, but more might be
                pass
            else:
                # Unused constants can be eliminated in any case.
                if not last_trace.getUsageCount():
                    if last_trace.getPrevious().isUnassignedTrace():
                        return (
                            None,
                            "new_statements",
                            "Dropped dead assignment statement to '%s'."
                            % (self.getVariableName()),
                        )
                    else:
                        return trace_collection.computedStatementResult(
                            statement=makeStatementDelVariable(
                                variable=self.variable,
                                version=self.variable_version,
                                tolerant=True,
                                source_ref=self.source_ref,
                            ),
                            change_tags="new_statements",
                            change_desc="Lowered dead assignment statement to '%s' to previous value 'del'."
                            % self.getVariableName(),
                        )

                # Still trace or assignment, for the last time. TODO: Maybe this can be
                # used for the keeping of the "replacement node" as well.
                self.variable_trace = trace_collection.onVariableSetToUnescapablePropagatedValue(
                    variable=variable,
                    version=self.variable_version,
                    assign_node=self,
                    replacement=lambda _replaced_node: self.subnode_source.makeClone(),
                )

                if last_trace.getPrevious().isUnassignedTrace():
                    result = None
                else:
                    result = makeStatementDelVariable(
                        variable=self.variable,
                        version=self.variable_version,
                        tolerant=True,
                        source_ref=self.source_ref,
                    )

                return (
                    result,
                    "new_statements",
                    "Dropped propagated assignment statement to '%s'."
                    % self.getVariableName(),
                )

        # Note: Keep this aligned with computeStatementAssignmentTraceUpdate
        self.variable_trace = trace_collection.onVariableSetToUnescapableValue(
            variable=variable, version=self.variable_version, assign_node=self
        )

        return self, None, None

    def computeStatementAssignmentTraceUpdate(self, trace_collection, source):
        self.variable_trace = trace_collection.onVariableSetToUnescapableValue(
            variable=self.variable, version=self.variable_version, assign_node=self
        )

    @staticmethod
    def hasVeryTrustedValue():
        """Does this assignment node have a very trusted value."""
        return False


class StatementAssignmentVariableConstantMutableTrusted(
    StatementAssignmentVariableConstantImmutable
):
    kind = "STATEMENT_ASSIGNMENT_VARIABLE_CONSTANT_MUTABLE_TRUSTED"

    @staticmethod
    def hasVeryTrustedValue():
        return False


class StatementAssignmentVariableConstantImmutableTrusted(
    StatementAssignmentVariableConstantImmutable
):
    kind = "STATEMENT_ASSIGNMENT_VARIABLE_CONSTANT_IMMUTABLE_TRUSTED"

    @staticmethod
    def hasVeryTrustedValue():
        return True


class StatementAssignmentVariableHardValue(
    StatementAssignmentVariableMixin, StatementAssignmentVariableHardValueBase
):
    kind = "STATEMENT_ASSIGNMENT_VARIABLE_HARD_VALUE"

    named_children = ("source",)
    nice_children = ("assignment source",)
    node_attributes = ("variable", "variable_version")
    auto_compute_handling = "post_init"

    __slots__ = (
        "variable_trace",
        # TODO: Does every variant have to care here
        "inplace_suspect",
    )

    # False alarm due to post_init, pylint: disable=attribute-defined-outside-init

    # TODO: Add parsing of node_attributes for init values, so we can avoid these too
    def postInitNode(self):
        self.variable_trace = None
        self.inplace_suspect = None

    @staticmethod
    def getReleaseEscape():
        return ControlFlowDescriptionNoEscape

    def computeStatement(self, trace_collection):
        # TODO: Way too ugly to have global trace kinds just here, and needs to
        # be abstracted somehow. But for now we let it live here.
        variable = self.variable

        # Let assignment source may re-compute first.
        source = trace_collection.onExpression(self.subnode_source)

        # No assignment will occur, if the assignment source raises, so give up
        # on this, and return it as the only side effect.
        if source.willRaiseAnyException():
            result = makeStatementExpressionOnlyReplacementNode(
                expression=source, node=self
            )

            del self.parent

            return (
                result,
                "new_raise",
                """\
Assignment raises exception in assigned value, removed assignment.""",
            )

        current_trace = trace_collection.getVariableCurrentTrace(variable)

        # If the variable was already assigned to the same value, we can drop this.
        if current_trace.isAssignTraceVeryTrusted() and (
            source.isExpressionImportModuleHard()
            or source.isExpressionImportModuleFixed()
        ):
            prev_assign_node = current_trace.getAssignNode()
            prev_source = prev_assign_node.subnode_source

            assert (
                prev_source.isExpressionImportModuleHard()
                or prev_source.isExpressionImportModuleFixed()
                or prev_source.isExpressionImportModuleNameHard()
            ), prev_source

            if prev_source.getValueName() == source.getValueName():
                trace_collection.updateVeryTrustedModuleVariable(
                    variable,
                    old_node=source,
                    new_node=prev_source,
                )

                return (
                    makeStatementExpressionOnlyReplacementNode(
                        expression=source, node=self
                    ),
                    "new_statements",
                    "Removed assignment of %s from itself which is known to be defined."
                    % variable.getDescription(),
                )

        # Note: Keep this aligned with computeStatementAssignmentTraceUpdate
        self.variable_trace = trace_collection.onVariableSetToVeryTrustedValue(
            variable=variable, version=self.variable_version, assign_node=self
        )

        return self, None, None

    def computeStatementAssignmentTraceUpdate(self, trace_collection, source):
        self.variable_trace = trace_collection.onVariableSetToVeryTrustedValue(
            variable=self.variable, version=self.variable_version, assign_node=self
        )

    @staticmethod
    def hasVeryTrustedValue():
        """Does this assignment node have a very trusted value."""
        return True


class StatementAssignmentVariableFromVariable(
    StatementAssignmentVariableMixin, StatementAssignmentVariableFromVariableBase
):
    kind = "STATEMENT_ASSIGNMENT_VARIABLE_FROM_VARIABLE"

    named_children = ("source",)
    nice_children = ("assignment source",)
    node_attributes = ("variable", "variable_version")
    auto_compute_handling = "post_init"

    __slots__ = (
        "variable_trace",
        # TODO: Does every variant have to care here
        "inplace_suspect",
    )

    # False alarm due to post_init, pylint: disable=attribute-defined-outside-init

    # TODO: Add parsing of node_attributes for init values, so we can avoid these too
    def postInitNode(self):
        self.variable_trace = None
        self.inplace_suspect = None

    @staticmethod
    def getReleaseEscape():
        # TODO: Variable type may know better.
        return ControlFlowDescriptionFullEscape

    def computeStatement(self, trace_collection):
        # TODO: Way too ugly to have global trace kinds just here, and needs to
        # be abstracted somehow. But for now we let it live here.
        old_source = self.subnode_source
        variable = self.variable

        # Assigning from and to the same variable, can be optimized away
        # immediately, there is no point in doing it. Exceptions are of course
        # module variables that collide with built-in names. TODO: In
        # specialization this could be considered right away as its own node
        # type, waiting for it to compute like this.
        if not variable.isModuleVariable() and old_source.getVariable() is variable:
            # A variable access that has a side effect, must be preserved,
            # so it can e.g. raise an exception, otherwise we can be fully
            # removed.
            if old_source.mayHaveSideEffects():
                result = makeStatementExpressionOnlyReplacementNode(
                    expression=old_source, node=self
                )

                result = trace_collection.onStatement(result)

                return (
                    result,
                    "new_statements",
                    """\
Lowered assignment of %s from itself to mere access of it."""
                    % variable.getDescription(),
                )
            else:
                return (
                    None,
                    "new_statements",
                    """\
Removed assignment of %s from itself which is known to be defined."""
                    % variable.getDescription(),
                )

        # Let assignment source may re-compute first.
        source = trace_collection.onExpression(self.subnode_source)

        if source is old_source:
            result = self, None, None
        else:
            if source.willRaiseAnyException():
                result = makeStatementExpressionOnlyReplacementNode(
                    expression=source, node=self
                )

                return (
                    result,
                    "new_raise",
                    """\
Assignment raises exception in assigned variable access, removed assignment.""",
                )

            result = self._considerSpecialization(old_source, source)
            result[0].parent = self.parent

        result[0].computeStatementAssignmentTraceUpdate(trace_collection, source)

        return result

    def computeStatementAssignmentTraceUpdate(self, trace_collection, source):
        if source.isExpressionVariableRef():
            self.variable_trace = trace_collection.onVariableSetAliasing(
                variable=self.variable,
                version=self.variable_version,
                assign_node=self,
                source=source,
            )
        else:
            # Set-up the trace to the trace collection, so future references will
            # find this assignment. TODO: We should for non-variables make sure we do
            # always specialize, since this is no longer a variable once it was
            # resolved.
            self.variable_trace = trace_collection.onVariableSet(
                variable=self.variable, version=self.variable_version, assign_node=self
            )

            # TODO: Determine from future use of assigned variable, if this is needed at all.
            trace_collection.removeKnowledge(source)

    @staticmethod
    def hasVeryTrustedValue():
        """Does this assignment node have a very trusted value."""
        return False


class StatementAssignmentVariableFromTempVariable(
    StatementAssignmentVariableMixin, StatementAssignmentVariableFromTempVariableBase
):
    kind = "STATEMENT_ASSIGNMENT_VARIABLE_FROM_TEMP_VARIABLE"

    named_children = ("source",)
    nice_children = ("assignment source",)
    node_attributes = ("variable", "variable_version")
    auto_compute_handling = "post_init"

    __slots__ = (
        "variable_trace",
        # TODO: Does every variant have to care here
        "inplace_suspect",
    )

    # False alarm due to post_init, pylint: disable=attribute-defined-outside-init

    # TODO: Add parsing of node_attributes for init values, so we can avoid these too
    def postInitNode(self):
        self.variable_trace = None
        self.inplace_suspect = None

    @staticmethod
    def getReleaseEscape():
        # TODO: Variable type may know better.
        return ControlFlowDescriptionFullEscape

    def computeStatement(self, trace_collection):
        # TODO: Way too ugly to have global trace kinds just here, and needs to
        # be abstracted somehow. But for now we let it live here.
        old_source = self.subnode_source
        variable = self.variable

        # Assigning from and to the same variable, can be optimized away
        # immediately, there is no point in doing it. Exceptions are of course
        # module variables that collide with built-in names.
        if old_source.getVariable() is variable:
            return (
                None,
                "new_statements",
                """\
Removed assignment of %s from itself which is known to be defined."""
                % variable.getDescription(),
            )

        # Let assignment source may re-compute first.
        source = trace_collection.onExpression(self.subnode_source)

        if source is old_source:
            result = self, None, None
        else:
            result = self._considerSpecialization(old_source, source)
            result[0].parent = self.parent

        result[0].computeStatementAssignmentTraceUpdate(trace_collection, source)

        return result

    def computeStatementAssignmentTraceUpdate(self, trace_collection, source):
        # Set-up the trace to the trace collection, so future references will
        # find this assignment.
        self.variable_trace = trace_collection.onVariableSet(
            variable=self.variable, version=self.variable_version, assign_node=self
        )

        # TODO: Determine from future use of assigned variable, if this is needed at all.
        trace_collection.removeKnowledge(source)

    @staticmethod
    def hasVeryTrustedValue():
        """Does this assignment node have a very trusted value."""
        return False


def makeStatementAssignmentVariableConstant(
    source, variable, variable_version, very_trusted, source_ref
):
    if source.isMutable():
        if very_trusted:
            return StatementAssignmentVariableConstantMutableTrusted(
                source=source,
                variable=variable,
                source_ref=source_ref,
                variable_version=variable_version,
            )
        else:
            return StatementAssignmentVariableConstantMutable(
                source=source,
                variable=variable,
                source_ref=source_ref,
                variable_version=variable_version,
            )
    else:
        if very_trusted:
            return StatementAssignmentVariableConstantImmutableTrusted(
                source=source,
                variable=variable,
                source_ref=source_ref,
                variable_version=variable_version,
            )
        else:
            return StatementAssignmentVariableConstantImmutable(
                source=source,
                variable=variable,
                source_ref=source_ref,
                variable_version=variable_version,
            )


def makeStatementAssignmentVariable(
    source, variable, source_ref, variable_version=None
):
    assert source is not None, source_ref

    if variable_version is None:
        variable_version = variable.allocateTargetNumber()

    if source.isCompileTimeConstant():
        return makeStatementAssignmentVariableConstant(
            source=source,
            variable=variable,
            variable_version=variable_version,
            very_trusted=False,
            source_ref=source_ref,
        )
    elif source.isExpressionVariableRef():
        return StatementAssignmentVariableFromVariable(
            source=source,
            variable=variable,
            variable_version=variable_version,
            source_ref=source_ref,
        )
    elif source.isExpressionTempVariableRef():
        return StatementAssignmentVariableFromTempVariable(
            source=source,
            variable=variable,
            variable_version=variable_version,
            source_ref=source_ref,
        )
    elif source.hasVeryTrustedValue():
        return StatementAssignmentVariableHardValue(
            source=source,
            variable=variable,
            variable_version=variable_version,
            source_ref=source_ref,
        )
    elif source.getTypeShape().isShapeIterator():
        return StatementAssignmentVariableIterator(
            source=source,
            variable=variable,
            variable_version=variable_version,
            source_ref=source_ref,
        )
    else:
        return StatementAssignmentVariableGeneric(
            source=source,
            variable=variable,
            variable_version=variable_version,
            source_ref=source_ref,
        )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#
# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Tool-specific initialization for the generic POSIX linker.

There normally shouldn't be any need to import this module directly.
It will usually be imported through the generic SCons.Tool.Tool()
selection method.

"""

import SCons.Tool
import SCons.Util
import SCons.Warnings
from SCons.Tool import createProgBuilder
from SCons.Tool.linkCommon import smart_link
from SCons.Tool.linkCommon.LoadableModule import setup_loadable_module_logic
from SCons.Tool.linkCommon.SharedLibrary import setup_shared_lib_logic


def generate(env):
    """Add Builders and construction variables for gnulink to an Environment."""
    createProgBuilder(env)

    setup_shared_lib_logic(env)
    setup_loadable_module_logic(env)

    env['SMARTLINK'] = smart_link
    env['LINK'] = "$SMARTLINK"
    env['LINKFLAGS'] = SCons.Util.CLVar('')

    # __RPATH is only set to something ($_RPATH typically) on platforms that support it.
    env['LINKCOM'] = '$LINK -o $TARGET $LINKFLAGS $__RPATH $SOURCES $_LIBDIRFLAGS $_LIBFLAGS'
    env['LIBDIRPREFIX'] = '-L'
    env['LIBDIRSUFFIX'] = ''
    env['_LIBFLAGS'] = '${_stripixes(LIBLINKPREFIX, LIBS, LIBLINKSUFFIX, LIBPREFIXES, LIBSUFFIXES, __env__, LIBLITERALPREFIX)}'
    env['LIBLINKPREFIX'] = '-l'
    env['LIBLINKSUFFIX'] = ''


def exists(env):
    # This module isn't really a Tool on its own, it's common logic for
    # other linkers.
    return None

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Serial and Parallel classes to execute build tasks.

The Jobs class provides a higher level interface to start,
stop, and wait on jobs.
"""

import SCons.compat

import logging
import os
import queue
import signal
import sys
import threading

from enum import Enum

import SCons.Errors
import SCons.Warnings


# The default stack size (in kilobytes) of the threads used to execute
# jobs in parallel.
#
# We use a stack size of 256 kilobytes. The default on some platforms
# is too large and prevents us from creating enough threads to fully
# parallelized the build. For example, the default stack size on linux
# is 8 MBytes.

explicit_stack_size = None
default_stack_size = 256

interrupt_msg = 'Build interrupted.'

class InterruptState:
    def __init__(self):
        self.interrupted = False

    def set(self):
        self.interrupted = True

    def __call__(self):
        return self.interrupted


class Jobs:
    """An instance of this class initializes N jobs, and provides
    methods for starting, stopping, and waiting on all N jobs.
    """

    def __init__(self, num, taskmaster):
        """
        Create 'num' jobs using the given taskmaster. The exact implementation
        used varies with the number of jobs requested and the state of the `legacy_sched` flag
        to `--experimental`.
        """

        # Importing GetOption here instead of at top of file to avoid
        # circular imports
        # pylint: disable=import-outside-toplevel
        from SCons.Script import GetOption

        stack_size = explicit_stack_size
        if stack_size is None:
            stack_size = default_stack_size

        experimental_option = GetOption('experimental') or []
        if 'legacy_sched' in experimental_option:
            if num > 1:
                self.job = LegacyParallel(taskmaster, num, stack_size)
            else:
                self.job = Serial(taskmaster)
        else:
            self.job = NewParallel(taskmaster, num, stack_size)

        self.num_jobs = num

    def run(self, postfunc=lambda: None):
        """Run the jobs.

        postfunc() will be invoked after the jobs has run. It will be
        invoked even if the jobs are interrupted by a keyboard
        interrupt (well, in fact by a signal such as either SIGINT,
        SIGTERM or SIGHUP). The execution of postfunc() is protected
        against keyboard interrupts and is guaranteed to run to
        completion."""
        self._setup_sig_handler()
        try:
            self.job.start()
        finally:
            postfunc()
            self._reset_sig_handler()

    def were_interrupted(self):
        """Returns whether the jobs were interrupted by a signal."""
        return self.job.interrupted()

    def _setup_sig_handler(self):
        """Setup an interrupt handler so that SCons can shutdown cleanly in
        various conditions:

          a) SIGINT: Keyboard interrupt
          b) SIGTERM: kill or system shutdown
          c) SIGHUP: Controlling shell exiting

        We handle all of these cases by stopping the taskmaster. It
        turns out that it's very difficult to stop the build process
        by throwing asynchronously an exception such as
        KeyboardInterrupt. For example, the python Condition
        variables (threading.Condition) and queues do not seem to be
        asynchronous-exception-safe. It would require adding a whole
        bunch of try/finally block and except KeyboardInterrupt all
        over the place.

        Note also that we have to be careful to handle the case when
        SCons forks before executing another process. In that case, we
        want the child to exit immediately.
        """
        def handler(signum, stack, self=self, parentpid=os.getpid()):
            if os.getpid() == parentpid:
                self.job.taskmaster.stop()
                self.job.interrupted.set()
            else:
                os._exit(2)  # pylint: disable=protected-access

        self.old_sigint = signal.signal(signal.SIGINT, handler)
        self.old_sigterm = signal.signal(signal.SIGTERM, handler)
        try:
            self.old_sighup = signal.signal(signal.SIGHUP, handler)
        except AttributeError:
            pass
        if (self.old_sigint is None) or (self.old_sigterm is None) or \
                (hasattr(self, "old_sighup") and self.old_sighup is None):
            msg = "Overwritting previous signal handler which was not installed from Python. " + \
                "Will not be able to reinstate and so will return to default handler."
            SCons.Warnings.warn(SCons.Warnings.SConsWarning, msg)

    def _reset_sig_handler(self):
        """Restore the signal handlers to their previous state (before the
         call to _setup_sig_handler()."""
        sigint_to_use = self.old_sigint if self.old_sigint is not None else signal.SIG_DFL
        sigterm_to_use = self.old_sigterm if self.old_sigterm is not None else signal.SIG_DFL
        signal.signal(signal.SIGINT, sigint_to_use)
        signal.signal(signal.SIGTERM, sigterm_to_use)
        try:
            sigterm_to_use = self.old_sighup if self.old_sighup is not None else signal.SIG_DFL
            signal.signal(signal.SIGHUP, sigterm_to_use)
        except AttributeError:
            pass


class Serial:
    """This class is used to execute tasks in series, and is more efficient
    than Parallel, but is only appropriate for non-parallel builds. Only
    one instance of this class should be in existence at a time.

    This class is not thread safe.
    """

    def __init__(self, taskmaster):
        """Create a new serial job given a taskmaster.

        The taskmaster's next_task() method should return the next task
        that needs to be executed, or None if there are no more tasks. The
        taskmaster's executed() method will be called for each task when it
        is successfully executed, or failed() will be called if it failed to
        execute (e.g. execute() raised an exception)."""

        self.taskmaster = taskmaster
        self.interrupted = InterruptState()

    def start(self):
        """Start the job. This will begin pulling tasks from the taskmaster
        and executing them, and return when there are no more tasks. If a task
        fails to execute (i.e. execute() raises an exception), then the job will
        stop."""

        while True:
            task = self.taskmaster.next_task()

            if task is None:
                break

            try:
                task.prepare()
                if task.needs_execute():
                    task.execute()
            except Exception:
                if self.interrupted():
                    try:
                        raise SCons.Errors.BuildError(
                            task.targets[0], errstr=interrupt_msg)
                    except Exception:
                        task.exception_set()
                else:
                    task.exception_set()

                # Let the failed() callback function arrange for the
                # build to stop if that's appropriate.
                task.failed()
            else:
                task.executed()

            task.postprocess()
        self.taskmaster.cleanup()


class Worker(threading.Thread):
    """A worker thread waits on a task to be posted to its request queue,
    dequeues the task, executes it, and posts a tuple including the task
    and a boolean indicating whether the task executed successfully. """

    def __init__(self, requestQueue, resultsQueue, interrupted):
        super().__init__()
        self.daemon = True
        self.requestQueue = requestQueue
        self.resultsQueue = resultsQueue
        self.interrupted = interrupted
        self.start()

    def run(self):
        while True:
            task = self.requestQueue.get()

            if task is None:
                # The "None" value is used as a sentinel by
                # ThreadPool.cleanup().  This indicates that there
                # are no more tasks, so we should quit.
                break

            try:
                if self.interrupted():
                    raise SCons.Errors.BuildError(
                        task.targets[0], errstr=interrupt_msg)
                task.execute()
            except Exception:
                task.exception_set()
                ok = False
            else:
                ok = True

            self.resultsQueue.put((task, ok))

class ThreadPool:
    """This class is responsible for spawning and managing worker threads."""

    def __init__(self, num, stack_size, interrupted):
        """Create the request and reply queues, and 'num' worker threads.

        One must specify the stack size of the worker threads. The
        stack size is specified in kilobytes.
        """
        self.requestQueue = queue.Queue(0)
        self.resultsQueue = queue.Queue(0)

        try:
            prev_size = threading.stack_size(stack_size * 1024)
        except RuntimeError as e:
            # Only print a warning if the stack size has been explicitly set.
            if explicit_stack_size is not None:
                msg = "Setting stack size is unsupported by this version of Python:\n    " + \
                    e.args[0]
                SCons.Warnings.warn(SCons.Warnings.StackSizeWarning, msg)
        except ValueError as e:
            msg = "Setting stack size failed:\n    " + str(e)
            SCons.Warnings.warn(SCons.Warnings.StackSizeWarning, msg)

        # Create worker threads
        self.workers = []
        for _ in range(num):
            worker = Worker(self.requestQueue, self.resultsQueue, interrupted)
            self.workers.append(worker)

        if 'prev_size' in locals():
            threading.stack_size(prev_size)

    def put(self, task):
        """Put task into request queue."""
        self.requestQueue.put(task)

    def get(self):
        """Remove and return a result tuple from the results queue."""
        return self.resultsQueue.get()

    def preparation_failed(self, task):
        self.resultsQueue.put((task, False))

    def cleanup(self):
        """
        Shuts down the thread pool, giving each worker thread a
        chance to shut down gracefully.
        """
        # For each worker thread, put a sentinel "None" value
        # on the requestQueue (indicating that there's no work
        # to be done) so that each worker thread will get one and
        # terminate gracefully.
        for _ in self.workers:
            self.requestQueue.put(None)

        # Wait for all of the workers to terminate.
        #
        # If we don't do this, later Python versions (2.4, 2.5) often
        # seem to raise exceptions during shutdown.  This happens
        # in requestQueue.get(), as an assertion failure that
        # requestQueue.not_full is notified while not acquired,
        # seemingly because the main thread has shut down (or is
        # in the process of doing so) while the workers are still
        # trying to pull sentinels off the requestQueue.
        #
        # Normally these terminations should happen fairly quickly,
        # but we'll stick a one-second timeout on here just in case
        # someone gets hung.
        for worker in self.workers:
            worker.join(1.0)
        self.workers = []

class LegacyParallel:
    """This class is used to execute tasks in parallel, and is somewhat
    less efficient than Serial, but is appropriate for parallel builds.

    This class is thread safe.
    """

    def __init__(self, taskmaster, num, stack_size):
        """Create a new parallel job given a taskmaster.

        The taskmaster's next_task() method should return the next
        task that needs to be executed, or None if there are no more
        tasks. The taskmaster's executed() method will be called
        for each task when it is successfully executed, or failed()
        will be called if the task failed to execute (i.e. execute()
        raised an exception).

        Note: calls to taskmaster are serialized, but calls to
        execute() on distinct tasks are not serialized, because
        that is the whole point of parallel jobs: they can execute
        multiple tasks simultaneously. """

        self.taskmaster = taskmaster
        self.interrupted = InterruptState()
        self.tp = ThreadPool(num, stack_size, self.interrupted)

        self.maxjobs = num

    def start(self):
        """Start the job. This will begin pulling tasks from the
        taskmaster and executing them, and return when there are no
        more tasks. If a task fails to execute (i.e. execute() raises
        an exception), then the job will stop."""

        jobs = 0

        while True:
            # Start up as many available tasks as we're
            # allowed to.
            while jobs < self.maxjobs:
                task = self.taskmaster.next_task()
                if task is None:
                    break

                try:
                    # prepare task for execution
                    task.prepare()
                except Exception:
                    task.exception_set()
                    task.failed()
                    task.postprocess()
                else:
                    if task.needs_execute():
                        # dispatch task
                        self.tp.put(task)
                        jobs += 1
                    else:
                        task.executed()
                        task.postprocess()

            if not task and not jobs:
                break

            # Let any/all completed tasks finish up before we go
            # back and put the next batch of tasks on the queue.
            while True:
                task, ok = self.tp.get()
                jobs -= 1

                if ok:
                    task.executed()
                else:
                    if self.interrupted():
                        try:
                            raise SCons.Errors.BuildError(
                                task.targets[0], errstr=interrupt_msg)
                        except Exception:
                            task.exception_set()

                    # Let the failed() callback function arrange
                    # for the build to stop if that's appropriate.
                    task.failed()

                task.postprocess()

                if self.tp.resultsQueue.empty():
                    break

        self.tp.cleanup()
        self.taskmaster.cleanup()

# An experimental new parallel scheduler that uses a leaders/followers pattern.
class NewParallel:

    class State(Enum):
        READY = 0
        SEARCHING = 1
        STALLED = 2
        COMPLETED = 3

    class Worker(threading.Thread):
        def __init__(self, owner):
            super().__init__()
            self.daemon = True
            self.owner = owner
            self.start()

        def run(self):
            self.owner._work()

    class FakeLock(object):
        def lock(self):
            pass
        def unlock(self):
            pass
        def __enter__(self):
            pass
        def __exit__(self, *args):
            pass

    class FakeCondition(object):
        def __init__(self, lock):
            pass
        def wait(self):
            fatal();
        def notify(self):
            pass
        def notify_all(self):
            pass
        def __enter__(self):
            pass
        def __exit__(self, *args):
            pass

    def __init__(self, taskmaster, num, stack_size):
        self.taskmaster = taskmaster
        self.max_workers = num
        self.stack_size = stack_size
        self.interrupted = InterruptState()
        self.workers = []

        # The `tm_lock` is what ensures that we only have one
        # thread interacting with the taskmaster at a time. It
        # also protects access to our state that gets updated
        # concurrently. The `can_search_cv` is associated with
        # this mutex.
        self.tm_lock = (threading.Lock if self.max_workers > 1 else NewParallel.FakeLock)()

        # Guarded under `tm_lock`.
        self.jobs = 0
        self.state = NewParallel.State.READY

        # The `can_search_cv` is used to manage a leader /
        # follower pattern for access to the taskmaster, and to
        # awaken from stalls.
        self.can_search_cv = (threading.Condition if self.max_workers > 1 else NewParallel.FakeCondition)(self.tm_lock)

        # The queue of tasks that have completed execution. The
        # next thread to obtain `tm_lock`` will retire them.
        self.results_queue_lock = (threading.Lock if self.max_workers > 1 else NewParallel.FakeLock)()
        self.results_queue = []

        if self.taskmaster.trace:
            self.trace = self._setup_logging()
        else:
            self.trace = False

    def _setup_logging(self):
        jl = logging.getLogger("Job")
        jl.setLevel(level=logging.DEBUG)
        jl.addHandler(self.taskmaster.trace.log_handler)
        return jl

    def trace_message(self, message):
        # This grabs the name of the function which calls trace_message()
        method_name = sys._getframe(1).f_code.co_name + "():"
        thread_id=threading.get_ident()
        self.trace.debug('%s.%s [Thread:%s] %s' % (type(self).__name__, method_name, thread_id, message))

    def start(self):
        if self.max_workers == 1:
            self._work()
        else:
            self._start_worker()
            while len(self.workers) > 0:
                self.workers[0].join()
                self.workers.pop(0)
        self.taskmaster.cleanup()

    def _maybe_start_worker(self):
        if self.max_workers > 1 and len(self.workers) < self.max_workers:
            if self.jobs >= len(self.workers):
                self._start_worker()

    def _start_worker(self):
        prev_size = self._adjust_stack_size()
        if self.trace:
            self.trace_message("Starting new worker thread")
        self.workers.append(NewParallel.Worker(self))
        self._restore_stack_size(prev_size)

    def _adjust_stack_size(self):
        try:
            prev_size = threading.stack_size(self.stack_size * 1024)
            return prev_size
        except AttributeError as e:
            # Only print a warning if the stack size has been
            # explicitly set.
            if explicit_stack_size is not None:
                msg = "Setting stack size is unsupported by this version of Python:\n    " + \
                    e.args[0]
                SCons.Warnings.warn(SCons.Warnings.StackSizeWarning, msg)
        except ValueError as e:
            msg = "Setting stack size failed:\n    " + str(e)
            SCons.Warnings.warn(SCons.Warnings.StackSizeWarning, msg)

        return None

    def _restore_stack_size(self, prev_size):
        if prev_size is not None:
            threading.stack_size(prev_size)

    def _work(self):

        task = None

        while True:

            # Obtain `tm_lock`, granting exclusive access to the taskmaster.
            with self.can_search_cv:

                if self.trace:
                    self.trace_message("Gained exclusive access")

                # Capture whether we got here with `task` set,
                # then drop our reference to the task as we are no
                # longer interested in the actual object.
                completed_task = (task is not None)
                task = None

                # We will only have `completed_task` set here if
                # we have looped back after executing a task. If
                # we have completed a task and find that we are
                # stalled, we should speculatively indicate that
                # we are no longer stalled by transitioning to the
                # 'ready' state which will bypass the condition
                # wait so that we immediately process the results
                # queue and hopefully light up new
                # work. Otherwise, stay stalled, and we will wait
                # in the condvar. Some other thread will come back
                # here with a completed task.
                if self.state == NewParallel.State.STALLED and completed_task:
                    if self.trace:
                        self.trace_message("Detected stall with completed task, bypassing wait")
                    self.state = NewParallel.State.READY

                # Wait until we are neither searching nor stalled.
                while self.state == NewParallel.State.SEARCHING or self.state == NewParallel.State.STALLED:
                    if self.trace:
                        self.trace_message("Search already in progress, waiting")
                    self.can_search_cv.wait()

                # If someone set the completed flag, bail.
                if self.state == NewParallel.State.COMPLETED:
                    if self.trace:
                        self.trace_message("Completion detected, breaking from main loop")
                    break

                # Set the searching flag to indicate that a thread
                # is currently in the critical section for
                # taskmaster work.
                #
                if self.trace:
                    self.trace_message("Starting search")
                self.state = NewParallel.State.SEARCHING

                # Bulk acquire the tasks in the results queue
                # under the result queue lock, then process them
                # all outside that lock. We need to process the
                # tasks in the results queue before looking for
                # new work because we might be unable to find new
                # work if we don't.
                results_queue = []
                with self.results_queue_lock:
                    results_queue, self.results_queue = self.results_queue, results_queue

                if self.trace:
                    self.trace_message(f"Found {len(results_queue)} completed tasks to process")
                for (rtask, rresult) in results_queue:
                    if rresult:
                        rtask.executed()
                    else:
                        if self.interrupted():
                            try:
                                raise SCons.Errors.BuildError(
                                    rtask.targets[0], errstr=interrupt_msg)
                            except Exception:
                                rtask.exception_set()

                        # Let the failed() callback function arrange
                        # for the build to stop if that's appropriate.
                        rtask.failed()

                    rtask.postprocess()
                    self.jobs -= 1

                # We are done with any task objects that were in
                # the results queue.
                results_queue.clear()

                # Now, turn the crank on the taskmaster until we
                # either run out of tasks, or find a task that
                # needs execution. If we run out of tasks, go idle
                # until results arrive if jobs are pending, or
                # mark the walk as complete if not.
                while self.state == NewParallel.State.SEARCHING:
                    if self.trace:
                        self.trace_message("Searching for new tasks")
                    task = self.taskmaster.next_task()

                    if task:
                        # We found a task. Walk it through the
                        # task lifecycle. If it does not need
                        # execution, just complete the task and
                        # look for the next one. Otherwise,
                        # indicate that we are no longer searching
                        # so we can drop out of this loop, execute
                        # the task outside the lock, and allow
                        # another thread in to search.
                        try:
                            task.prepare()
                        except Exception:
                            task.exception_set()
                            task.failed()
                            task.postprocess()
                        else:
                            if not task.needs_execute():
                                if self.trace:
                                    self.trace_message("Found internal task")
                                task.executed()
                                task.postprocess()
                            else:
                                self.jobs += 1
                                if self.trace:
                                    self.trace_message("Found task requiring execution")
                                self.state = NewParallel.State.READY
                                self.can_search_cv.notify()
                                # This thread will be busy taking care of
                                # `execute`ing this task. If we haven't
                                # reached the limit, spawn a new thread to
                                # turn the crank and find the next task.
                                self._maybe_start_worker()

                    else:
                        # We failed to find a task, so this thread
                        # cannot continue turning the taskmaster
                        # crank. We must exit the loop.
                        if self.jobs:
                            # No task was found, but there are
                            # outstanding jobs executing that
                            # might unblock new tasks when they
                            # complete. Transition to the stalled
                            # state. We do not need a notify,
                            # because we know there are threads
                            # outstanding that will re-enter the
                            # loop.
                            #
                            if self.trace:
                                self.trace_message("Found no task requiring execution, but have jobs: marking stalled")
                            self.state = NewParallel.State.STALLED
                        else:
                            # We didn't find a task and there are
                            # no jobs outstanding, so there is
                            # nothing that will ever return
                            # results which might unblock new
                            # tasks. We can conclude that the walk
                            # is complete. Update our state to
                            # note completion and awaken anyone
                            # sleeping on the condvar.
                            #
                            if self.trace:
                                self.trace_message("Found no task requiring execution, and have no jobs: marking complete")
                            self.state = NewParallel.State.COMPLETED
                            self.can_search_cv.notify_all()

            # We no longer hold `tm_lock` here. If we have a task,
            # we can now execute it. If there are threads waiting
            # to search, one of them can now begin turning the
            # taskmaster crank in NewParallel.
            if task:
                if self.trace:
                    self.trace_message("Executing task")
                ok = True
                try:
                    if self.interrupted():
                        raise SCons.Errors.BuildError(
                            task.targets[0], errstr=interrupt_msg)
                    task.execute()
                except Exception:
                    ok = False
                    task.exception_set()

                # Grab the results queue lock and enqueue the
                # executed task and state. The next thread into
                # the searching loop will complete the
                # postprocessing work under the taskmaster lock.
                #
                if self.trace:
                    self.trace_message("Enqueueing executed task results")
                with self.results_queue_lock:
                    self.results_queue.append((task, ok))

            # Tricky state "fallthrough" here. We are going back
            # to the top of the loop, which behaves differently
            # depending on whether `task` is set. Do not perturb
            # the value of the `task` variable if you add new code
            # after this comment.

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Code generation for slicing.

This is about slice lookups, assignments, and deletions. There is also a
special case, for using index values instead of objects. The slice objects
are also created here, and can be used for indexing.
"""

from nuitka.PythonVersions import python_version
from nuitka.States import states

from .CodeHelpers import (
    generateChildExpressionsCode,
    generateExpressionCode,
    generateExpressionsCode,
    withObjectCodeTemporaryAssignment,
)
from .ErrorCodes import getErrorExitBoolCode, getErrorExitCode
from .IndexCodes import (
    getIndexCode,
    getIndexValueCode,
    getMaxIndexCode,
    getMinIndexCode,
)


def _isSmallNumberConstant(node):
    if node.isNumberConstant():
        value = node.getCompileTimeConstant()
        return abs(int(value)) < 2**63 - 1
    else:
        return False


def _generateSliceRangeIdentifier(lower, upper, scope, emit, context):
    lower_name = context.allocateTempName(scope + "slicedel_index_lower", "Py_ssize_t")
    upper_name = context.allocateTempName(scope + "_index_upper", "Py_ssize_t")

    if lower is None:
        getMinIndexCode(to_name=lower_name, emit=emit)
    elif lower.isExpressionConstantRef() and _isSmallNumberConstant(lower):
        getIndexValueCode(
            to_name=lower_name, value=int(lower.getCompileTimeConstant()), emit=emit
        )
    else:
        value_name = context.allocateTempName(scope + "_lower_index_value")

        generateExpressionCode(
            to_name=value_name, expression=lower, emit=emit, context=context
        )

        getIndexCode(
            to_name=lower_name, value_name=value_name, emit=emit, context=context
        )

    if upper is None:
        getMaxIndexCode(to_name=upper_name, emit=emit)
    elif upper.isExpressionConstantRef() and _isSmallNumberConstant(upper):
        getIndexValueCode(
            to_name=upper_name, value=int(upper.getCompileTimeConstant()), emit=emit
        )
    else:
        value_name = context.allocateTempName(scope + "_upper_index_value")

        generateExpressionCode(
            to_name=value_name, expression=upper, emit=emit, context=context
        )

        getIndexCode(
            to_name=upper_name, value_name=value_name, emit=emit, context=context
        )

    return lower_name, upper_name


def _decideSlicing(lower, upper):
    return (lower is None or lower.isIndexable()) and (
        upper is None or upper.isIndexable()
    )


def generateSliceLookupCode(to_name, expression, emit, context):
    assert python_version < 0x300

    lower = expression.subnode_lower
    upper = expression.subnode_upper

    if _decideSlicing(lower, upper):
        lower_name, upper_name = _generateSliceRangeIdentifier(
            lower=lower, upper=upper, scope="slice", emit=emit, context=context
        )

        source_name = context.allocateTempName("slice_source")

        generateExpressionCode(
            to_name=source_name,
            expression=expression.subnode_expression,
            emit=emit,
            context=context,
        )

        with withObjectCodeTemporaryAssignment(
            to_name, "slice_result", expression, emit, context
        ) as result_name:
            _getSliceLookupIndexesCode(
                to_name=result_name,
                source_name=source_name,
                lower_name=lower_name,
                upper_name=upper_name,
                emit=emit,
                context=context,
            )
    else:
        source_name, lower_name, upper_name = generateExpressionsCode(
            names=("slice_source", "slice_lower", "slice_upper"),
            expressions=(
                expression.subnode_expression,
                expression.subnode_lower,
                expression.subnode_upper,
            ),
            emit=emit,
            context=context,
        )

        with withObjectCodeTemporaryAssignment(
            to_name, "slice_result", expression, emit, context
        ) as result_name:
            _getSliceLookupCode(
                to_name=result_name,
                source_name=source_name,
                lower_name=lower_name,
                upper_name=upper_name,
                emit=emit,
                context=context,
            )


def generateAssignmentSliceCode(statement, emit, context):
    assert python_version < 0x300

    lookup_source = statement.subnode_expression
    lower = statement.subnode_lower
    upper = statement.subnode_upper
    value = statement.subnode_source

    value_name = context.allocateTempName("sliceass_value")

    generateExpressionCode(
        to_name=value_name, expression=value, emit=emit, context=context
    )

    target_name = context.allocateTempName("sliceass_target")

    generateExpressionCode(
        to_name=target_name, expression=lookup_source, emit=emit, context=context
    )

    if _decideSlicing(lower, upper):
        lower_name, upper_name = _generateSliceRangeIdentifier(
            lower=lower, upper=upper, scope="sliceass", emit=emit, context=context
        )

        with context.withCurrentSourceCodeReference(
            value.getSourceReference()
            if states.is_full_compat
            else statement.getSourceReference()
        ):
            _getSliceAssignmentIndexesCode(
                target_name=target_name,
                lower_name=lower_name,
                upper_name=upper_name,
                value_name=value_name,
                emit=emit,
                context=context,
            )
    else:
        lower_name, upper_name = generateExpressionsCode(
            names=("sliceass_lower", "sliceass_upper"),
            expressions=(lower, upper),
            emit=emit,
            context=context,
        )

        with context.withCurrentSourceCodeReference(
            value.getSourceReference()
            if states.is_full_compat
            else statement.getSourceReference()
        ):
            _getSliceAssignmentCode(
                target_name=target_name,
                upper_name=upper_name,
                lower_name=lower_name,
                value_name=value_name,
                emit=emit,
                context=context,
            )


def generateDelSliceCode(statement, emit, context):
    assert python_version < 0x300

    target = statement.subnode_expression
    lower = statement.subnode_lower
    upper = statement.subnode_upper

    target_name = context.allocateTempName("slicedel_target")

    generateExpressionCode(
        to_name=target_name, expression=target, emit=emit, context=context
    )

    if _decideSlicing(lower, upper):
        lower_name, upper_name = _generateSliceRangeIdentifier(
            lower=lower, upper=upper, scope="slicedel", emit=emit, context=context
        )

        with context.withCurrentSourceCodeReference(
            (upper or lower or statement).getSourceReference()
            if states.is_full_compat
            else statement.getSourceReference()
        ):
            _getSliceDelIndexesCode(
                target_name=target_name,
                lower_name=lower_name,
                upper_name=upper_name,
                emit=emit,
                context=context,
            )
    else:
        lower_name, upper_name = generateExpressionsCode(
            names=("slicedel_lower", "slicedel_upper"),
            expressions=(lower, upper),
            emit=emit,
            context=context,
        )

        with context.withCurrentSourceCodeReference(
            (upper or lower or target).getSourceReference()
            if states.is_full_compat
            else statement.getSourceReference()
        ):
            _getSliceDelCode(
                target_name=target_name,
                lower_name=lower_name,
                upper_name=upper_name,
                emit=emit,
                context=context,
            )


def generateBuiltinSlice3Code(to_name, expression, emit, context):
    lower_name, upper_name, step_name = generateChildExpressionsCode(
        expression=expression, emit=emit, context=context
    )

    with withObjectCodeTemporaryAssignment(
        to_name, "slice_obj_value", expression, emit, context
    ) as result_name:
        emit(
            "%s = MAKE_SLICE_OBJECT3(tstate, %s, %s, %s);"
            % (
                result_name,
                lower_name,
                upper_name,
                step_name,
            )
        )

        getErrorExitCode(
            check_name=result_name,
            release_names=(lower_name, upper_name, step_name),
            needs_check=False,  # Note: Cannot fail
            emit=emit,
            context=context,
        )

        context.addCleanupTempName(result_name)


def generateBuiltinSlice2Code(to_name, expression, emit, context):
    lower_name, upper_name = generateChildExpressionsCode(
        expression=expression, emit=emit, context=context
    )

    with withObjectCodeTemporaryAssignment(
        to_name, "slice_obj_value", expression, emit, context
    ) as result_name:
        emit(
            "%s = MAKE_SLICE_OBJECT2(tstate, %s, %s);"
            % (
                result_name,
                lower_name,
                upper_name,
            )
        )

        getErrorExitCode(
            check_name=result_name,
            release_names=(lower_name, upper_name),
            needs_check=False,  # Note: Cannot fail
            emit=emit,
            context=context,
        )

        context.addCleanupTempName(result_name)


def generateBuiltinSlice1Code(to_name, expression, emit, context):
    (upper_name,) = generateChildExpressionsCode(
        expression=expression, emit=emit, context=context
    )

    with withObjectCodeTemporaryAssignment(
        to_name, "slice_obj_value", expression, emit, context
    ) as result_name:
        emit(
            "%s = MAKE_SLICE_OBJECT1(tstate, %s);"
            % (
                result_name,
                upper_name,
            )
        )

        getErrorExitCode(
            check_name=result_name,
            release_name=upper_name,
            needs_check=False,  # Note: Cannot fail
            emit=emit,
            context=context,
        )

        context.addCleanupTempName(result_name)


def _getSliceLookupCode(to_name, source_name, lower_name, upper_name, emit, context):
    emit(
        "%s = LOOKUP_SLICE(tstate, %s, %s, %s);"
        % (
            to_name,
            source_name,
            lower_name if lower_name is not None else "Py_None",
            upper_name if upper_name is not None else "Py_None",
        )
    )

    getErrorExitCode(
        check_name=to_name,
        release_names=(source_name, lower_name, upper_name),
        emit=emit,
        context=context,
    )

    context.addCleanupTempName(to_name)


def _getSliceLookupIndexesCode(
    to_name, lower_name, upper_name, source_name, emit, context
):
    emit(
        "%s = LOOKUP_INDEX_SLICE(%s, %s, %s);"
        % (to_name, source_name, lower_name, upper_name)
    )

    getErrorExitCode(
        check_name=to_name, release_name=source_name, emit=emit, context=context
    )

    context.addCleanupTempName(to_name)


def _getSliceAssignmentIndexesCode(
    target_name, lower_name, upper_name, value_name, emit, context
):
    res_name = context.getBoolResName()

    emit(
        "%s = SET_INDEX_SLICE(%s, %s, %s, %s);"
        % (res_name, target_name, lower_name, upper_name, value_name)
    )

    getErrorExitBoolCode(
        condition="%s == false" % res_name,
        release_names=(value_name, target_name),
        emit=emit,
        context=context,
    )


def _getSliceAssignmentCode(
    target_name, lower_name, upper_name, value_name, emit, context
):
    res_name = context.getBoolResName()

    emit(
        "%s = SET_SLICE(tstate, %s, %s, %s, %s);"
        % (
            res_name,
            target_name,
            lower_name if lower_name is not None else "Py_None",
            upper_name if upper_name is not None else "Py_None",
            value_name,
        )
    )

    getErrorExitBoolCode(
        condition="%s == false" % res_name,
        release_names=(target_name, lower_name, upper_name, value_name),
        emit=emit,
        context=context,
    )


def _getSliceDelIndexesCode(target_name, lower_name, upper_name, emit, context):
    res_name = context.getBoolResName()

    emit(
        "%s = DEL_INDEX_SLICE(%s, %s, %s);"
        % (res_name, target_name, lower_name, upper_name)
    )

    getErrorExitBoolCode(
        condition="%s == false" % res_name,
        release_name=target_name,
        emit=emit,
        context=context,
    )


def _getSliceDelCode(target_name, lower_name, upper_name, emit, context):
    res_name = context.getBoolResName()

    emit(
        "%s = DEL_SLICE(tstate, %s, %s, %s);"
        % (
            res_name,
            target_name,
            lower_name if lower_name is not None else "Py_None",
            upper_name if upper_name is not None else "Py_None",
        )
    )

    getErrorExitBoolCode(
        condition="%s == false" % res_name,
        release_names=(target_name, lower_name, upper_name),
        emit=emit,
        context=context,
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Runner for package tests of Nuitka.

Package tests are typically aiming at checking specific module constellations
in module mode and making sure the details are being right there. These are
synthetic small packages, each of which try to demonstrate one or more points
or special behavior.

"""


import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

from nuitka.tools.testing.Common import (
    compareWithCPython,
    createSearchMode,
    getTempDir,
    my_print,
    scanDirectoryForTestCaseFolders,
    setup,
)
from nuitka.utils.FileOperations import getSubDirectories, relpath


def main():
    setup(suite="packages")

    search_mode = createSearchMode()

    for filename, filename_main in scanDirectoryForTestCaseFolders(
        ".", allow_none=True
    ):
        active = search_mode.consider(dirname=None, filename=filename)

        if filename_main is None:
            filename_main = relpath(path=getSubDirectories(filename)[0], start=filename)

        if active:
            my_print("Consider output of compiled package:", filename)

            extra_flags = [
                "--mode=module",
                "expect_success",
                "remove_output",
                "two_step_execution",
            ]

            extra_flags.append("--output-dir=%s" % getTempDir())

            # TODO: This mismatches a rename and means the test is never
            # actually ran.
            if filename == "top_level_attributes":
                extra_flags.append("--mode=module-entry-point=main")

            compareWithCPython(
                dirname=filename,
                filename=filename_main,
                extra_flags=extra_flags,
                search_mode=search_mode,
            )

    search_mode.finish()


if __name__ == "__main__":
    main()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Code generation for implicit and explicit exception raises.

Exceptions from other operations are consider ErrorCodes domain.

"""

from nuitka.Builtins import isBaseExceptionSimpleExtension
from nuitka.PythonVersions import python_version
from nuitka.States import states

from .CodeHelpers import (
    generateChildExpressionsCode,
    generateExpressionCode,
    withObjectCodeTemporaryAssignment,
)
from .ErrorCodes import getErrorExitCode, getFrameVariableTypeDescriptionCode
from .ExceptionCodes import getExceptionIdentifier
from .LabelCodes import getGotoCode
from .LineNumberCodes import (
    emitErrorLineNumberUpdateCode,
    getErrorLineNumberUpdateCode,
)
from .PythonAPICodes import getReferenceExportCode


def generateReraiseCode(statement, emit, context):
    with context.withCurrentSourceCodeReference(
        value=statement.getCompatibleSourceReference()
    ):
        getReRaiseExceptionCode(emit=emit, context=context)


def _haveQuickExceptionCreationCode(exception_name):
    result = exception_name in ("StopIteration",)

    # TODO: Get this complete for larger test suites.
    # assert result, exception_name

    return result


def _generateExceptionNormalizeCode(to_name, exception_type, emit, context):

    if exception_type.isExpressionBuiltinExceptionRef():
        exception_name = exception_type.getExceptionName()
        if isBaseExceptionSimpleExtension(exception_type.getCompileTimeConstant()):
            emit(
                "%s = MAKE_BASE_EXCEPTION_DERIVED_EMPTY(%s);"
                % (to_name, getExceptionIdentifier(exception_name))
            )
            context.addCleanupTempName(to_name)

            return
        elif _haveQuickExceptionCreationCode(exception_type.getExceptionName()):
            # TODO: Code generation like this here, ought to be shared with
            # the one for ExpressionBuiltinMakeExceptionRef once that is
            # done too.
            if exception_name == "StopIteration":
                emit("%s = MAKE_STOP_ITERATION_EMPTY();" % to_name)
            else:
                assert False, exception_name

            context.addCleanupTempName(to_name)
            return

    # TODO: Can only fall through if exception builtins are not fully handled,
    # but we should make sure that doesn't ever happen by handling them all.

    if exception_type.isExpressionBuiltinMakeException():
        # TODO: Make this shape based, but currently the nodes do not
        # provide the shape, we need to add that.

        # No need to normalize these.
        generateExpressionCode(
            to_name=to_name,
            expression=exception_type,
            emit=emit,
            context=context,
        )
    else:
        # Need to normalize exceptions.
        raise_type_input_name = context.allocateTempName("raise_type_input")

        generateExpressionCode(
            to_name=raise_type_input_name,
            expression=exception_type,
            emit=emit,
            context=context,
        )

        emit(
            "%s = NORMALIZE_EXCEPTION_VALUE_FOR_RAISE(tstate, %s);"
            % (to_name, raise_type_input_name)
        )

        getErrorExitCode(
            check_name=to_name,
            release_name=raise_type_input_name,
            needs_check=True,
            emit=emit,
            context=context,
        )

        context.addCleanupTempName(to_name)


def generateRaiseCode(statement, emit, context):
    exception_type = statement.subnode_exception_type
    exception_value = statement.subnode_exception_value
    exception_tb = statement.subnode_exception_trace
    exception_cause = statement.subnode_exception_cause

    # Exception cause is only possible with simple raise form.
    if exception_cause is not None:
        assert exception_type is not None
        assert exception_value is None
        assert exception_tb is None

        raise_type_name = context.allocateTempName("raise_type")

        if python_version >= 0x3C0:
            _generateExceptionNormalizeCode(
                to_name=raise_type_name,
                exception_type=exception_type,
                emit=emit,
                context=context,
            )
        else:
            generateExpressionCode(
                to_name=raise_type_name,
                expression=exception_type,
                emit=emit,
                context=context,
            )

        raise_cause_name = context.allocateTempName("raise_cause")

        generateExpressionCode(
            to_name=raise_cause_name,
            expression=exception_cause,
            emit=emit,
            context=context,
        )

        with context.withCurrentSourceCodeReference(
            exception_cause.getSourceReference()
        ):
            _getRaiseExceptionWithCauseCode(
                raise_type_name=raise_type_name,
                raise_cause_name=raise_cause_name,
                emit=emit,
                context=context,
            )
    elif exception_type is None:
        assert False, statement
    elif exception_value is None and exception_tb is None:
        raise_type_name = context.allocateTempName("raise_type")

        if python_version >= 0x3C0:
            _generateExceptionNormalizeCode(
                to_name=raise_type_name,
                exception_type=exception_type,
                emit=emit,
                context=context,
            )
        else:
            generateExpressionCode(
                to_name=raise_type_name,
                expression=exception_type,
                emit=emit,
                context=context,
            )

        with context.withCurrentSourceCodeReference(
            value=exception_type.getCompatibleSourceReference()
        ):
            _getRaiseExceptionWithTypeCode(
                raise_type_name=raise_type_name, emit=emit, context=context
            )

    elif exception_tb is None:
        raise_type_name = context.allocateTempName("raise_type")

        generateExpressionCode(
            to_name=raise_type_name,
            expression=exception_type,
            emit=emit,
            context=context,
        )

        raise_value_name = context.allocateTempName("raise_value")

        generateExpressionCode(
            to_name=raise_value_name,
            expression=exception_value,
            emit=emit,
            context=context,
        )

        with context.withCurrentSourceCodeReference(
            exception_value.getCompatibleSourceReference()
        ):
            _getRaiseExceptionWithValueCode(
                raise_type_name=raise_type_name,
                raise_value_name=raise_value_name,
                emit=emit,
                context=context,
            )
    else:
        raise_type_name = context.allocateTempName("raise_type")

        generateExpressionCode(
            to_name=raise_type_name,
            expression=exception_type,
            emit=emit,
            context=context,
        )

        raise_value_name = context.allocateTempName("raise_value")

        generateExpressionCode(
            to_name=raise_value_name,
            expression=exception_value,
            emit=emit,
            context=context,
        )

        raise_tb_name = context.allocateTempName("raise_tb")

        generateExpressionCode(
            to_name=raise_tb_name, expression=exception_tb, emit=emit, context=context
        )

        with context.withCurrentSourceCodeReference(exception_tb.getSourceReference()):
            _getRaiseExceptionWithTracebackCode(
                raise_type_name=raise_type_name,
                raise_value_name=raise_value_name,
                raise_tb_name=raise_tb_name,
                emit=emit,
                context=context,
            )


def generateRaiseExpressionCode(to_name, expression, emit, context):
    (exception_value_name,) = generateChildExpressionsCode(
        expression=expression, emit=emit, context=context
    )

    # Missed optimization opportunity, please report it, this should not
    # normally happen. We are supposed to propagate this upwards.
    if states.is_debug:
        # TODO: Need to optimize ExpressionLocalsVariableRefOrFallback once we know
        # it handles cases where the value is not in locals dict properly.

        parent = expression.parent
        assert (
            parent.isExpressionSideEffects()
            or parent.isExpressionConditional()
            or parent.isExpressionConditionalOr()
            or parent.isExpressionConditionalAnd()
            or parent.isExpressionLocalsVariableRefOrFallback()
        ), (expression, expression.parent, expression.asXmlText())

    with withObjectCodeTemporaryAssignment(
        to_name, "raise_exception_result", expression, emit, context
    ) as value_name:
        # That's how we indicate exception to the surrounding world.
        emit("%s = NULL;" % value_name)

        _getRaiseExceptionWithTypeCode(
            raise_type_name=exception_value_name,
            emit=emit,
            context=context,
        )


def getReRaiseExceptionCode(emit, context):
    (
        exception_state_name,
        exception_lineno,
    ) = context.variable_storage.getExceptionVariableDescriptions()

    (
        keeper_exception_state_name,
        _keeper_lineno,
    ) = context.getExceptionKeeperVariables()

    if keeper_exception_state_name is None:
        emit(
            """\
%(bool_res_name)s = RERAISE_EXCEPTION(tstate, &%(exception_state_name)s);
if (unlikely(%(bool_res_name)s == false)) {
    %(update_code)s
}
"""
            % {
                "exception_state_name": exception_state_name,
                "bool_res_name": context.getBoolResName(),
                "update_code": getErrorLineNumberUpdateCode(context),
            }
        )

        frame_handle = context.getFrameHandle()

        if frame_handle:
            emit(
                """\
{
    PyTracebackObject *exception_tb = GET_EXCEPTION_STATE_TRACEBACK(&%(exception_state_name)s);

    if ((exception_tb != NULL) && (exception_tb->tb_frame == &%(frame_identifier)s->m_frame)) {
        %(frame_identifier)s->m_frame.f_lineno = exception_tb->tb_lineno;
    }
}"""
                % {
                    "exception_state_name": exception_state_name,
                    "frame_identifier": context.getFrameHandle(),
                }
            )

            emit(getFrameVariableTypeDescriptionCode(context))
    else:
        (
            keeper_exception_state_name,
            keeper_lineno,
        ) = context.getExceptionKeeperVariables()

        emit(
            """\
// Re-raise.
%(exception_state_name)s = %(keeper_exception_state_name)s;
%(exception_lineno)s = %(keeper_lineno)s;
"""
            % {
                "exception_state_name": exception_state_name,
                "exception_lineno": exception_lineno,
                "keeper_exception_state_name": keeper_exception_state_name,
                "keeper_lineno": keeper_lineno,
            }
        )

    getGotoCode(context.getExceptionEscape(), emit)


def _getRaiseExceptionWithCauseCode(raise_type_name, raise_cause_name, emit, context):
    (
        exception_state_name,
        _exception_lineno,
    ) = context.variable_storage.getExceptionVariableDescriptions()

    if python_version < 0x3C0:
        emit("%s.exception_type = %s;" % (exception_state_name, raise_type_name))
    else:
        emit("%s.exception_value = %s;" % (exception_state_name, raise_type_name))

    getReferenceExportCode(raise_type_name, emit, context)

    getReferenceExportCode(raise_cause_name, emit, context)

    emitErrorLineNumberUpdateCode(emit, context)
    emit(
        "RAISE_EXCEPTION_WITH_CAUSE(tstate, &%s, %s);"
        % (exception_state_name, raise_cause_name)
    )

    emit(getFrameVariableTypeDescriptionCode(context))

    getGotoCode(context.getExceptionEscape(), emit)

    if context.needsCleanup(raise_type_name):
        context.removeCleanupTempName(raise_type_name)
    if context.needsCleanup(raise_cause_name):
        context.removeCleanupTempName(raise_cause_name)


def _getRaiseExceptionWithTypeCode(raise_type_name, emit, context):
    (
        exception_state_name,
        _exception_lineno,
    ) = context.variable_storage.getExceptionVariableDescriptions()

    if python_version < 0x3C0:
        emit("%s.exception_type = %s;" % (exception_state_name, raise_type_name))
        getReferenceExportCode(raise_type_name, emit, context)

        emitErrorLineNumberUpdateCode(emit, context)

        emit("RAISE_EXCEPTION_WITH_TYPE(tstate, &%s);" % exception_state_name)
    else:
        emit("%s.exception_value = %s;" % (exception_state_name, raise_type_name))
        getReferenceExportCode(raise_type_name, emit, context)

        emitErrorLineNumberUpdateCode(emit, context)

        emit("RAISE_EXCEPTION_WITH_VALUE(tstate, &%s);" % exception_state_name)

    emit(getFrameVariableTypeDescriptionCode(context))

    getGotoCode(context.getExceptionEscape(), emit)

    if context.needsCleanup(raise_type_name):
        context.removeCleanupTempName(raise_type_name)


def _getRaiseExceptionWithValueCode(raise_type_name, raise_value_name, emit, context):
    (
        exception_state_name,
        _exception_lineno,
    ) = context.variable_storage.getExceptionVariableDescriptions()

    emit("%s.exception_type = %s;" % (exception_state_name, raise_type_name))
    getReferenceExportCode(raise_type_name, emit, context)
    emit("%s.exception_value = %s;" % (exception_state_name, raise_value_name))
    getReferenceExportCode(raise_value_name, emit, context)

    emitErrorLineNumberUpdateCode(emit, context)

    emit("RAISE_EXCEPTION_WITH_TYPE_AND_VALUE(tstate, &%s);" % (exception_state_name,))

    emit(getFrameVariableTypeDescriptionCode(context))

    getGotoCode(context.getExceptionEscape(), emit)

    if context.needsCleanup(raise_type_name):
        context.removeCleanupTempName(raise_type_name)
    if context.needsCleanup(raise_value_name):
        context.removeCleanupTempName(raise_value_name)


def _getRaiseExceptionWithTracebackCode(
    raise_type_name, raise_value_name, raise_tb_name, emit, context
):
    (
        exception_state_name,
        _exception_lineno,
    ) = context.variable_storage.getExceptionVariableDescriptions()

    emit("%s.exception_type = %s;" % (exception_state_name, raise_type_name))
    getReferenceExportCode(raise_type_name, emit, context)
    emit("%s.exception_value = %s;" % (exception_state_name, raise_value_name))
    getReferenceExportCode(raise_value_name, emit, context)
    emit(
        "%s.exception_tb = (PyTracebackObject *)%s;"
        % (exception_state_name, raise_tb_name)
    )
    getReferenceExportCode(raise_tb_name, emit, context)

    emit("RAISE_EXCEPTION_WITH_TRACEBACK(tstate, &%s);" % (exception_state_name))

    # If anything is wrong, that will be used.
    emitErrorLineNumberUpdateCode(emit, context)

    emit(getFrameVariableTypeDescriptionCode(context))

    getGotoCode(context.getExceptionEscape(), emit)

    if context.needsCleanup(raise_type_name):
        context.removeCleanupTempName(raise_type_name)
    if context.needsCleanup(raise_value_name):
        context.removeCleanupTempName(raise_value_name)
    if context.needsCleanup(raise_tb_name):
        context.removeCleanupTempName(raise_tb_name)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Functions for Microsoft Visual C/C++.

The _reset method is used to restore MSVC module data structures to their
initial state for testing purposes.

The _verify method is used as a sanity check that MSVC module data structures
are internally consistent.

Currently:
* _reset is invoked from reset_installed_vcs in the vc module.
* _verify is invoked from the last line in the vc module.
"""

from . import Exceptions  # noqa: F401

from . import Config  # noqa: F401
from . import Util  # noqa: F401
from . import Registry  # noqa: F401
from . import Kind  # noqa: F401
from . import SetupEnvDefault  # noqa: F401
from . import Policy  # noqa: F401
from . import WinSDK  # noqa: F401
from . import ScriptArguments  # noqa: F401

from . import Dispatcher as _Dispatcher

def _reset():
    _Dispatcher.reset()

def _verify():
    _Dispatcher.verify()




#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Pack and copy files for standalone mode.

This is expected to work for macOS, Windows, and Linux. Other things like
FreeBSD are also very welcome, but might break with time and need your
help.
"""

import os

from nuitka.containers.OrderedSets import OrderedSet
from nuitka.Errors import NuitkaForbiddenDLLEncounter, NuitkaNotYetSupported
from nuitka.importing.Importing import (
    getPythonUnpackedSearchPath,
    locateModule,
)
from nuitka.importing.StandardLibrary import isStandardLibraryPath
from nuitka.options.Options import (
    isShowProgress,
    shallCreateAppBundle,
    shallNotStoreDependsExeCachedResults,
    shallNotUseDependsExeCachedResults,
)
from nuitka.plugins.Hooks import (
    decideAllowOutsideDependencies,
    onCopiedDLLs,
    removeDllDependencies,
)
from nuitka.Progress import (
    closeProgressBar,
    reportProgressBar,
    setupProgressBar,
)
from nuitka.PythonFlavors import (
    getHomebrewInstallPath,
    isAnacondaPython,
    isCPythonOfficialPackage,
    isHomebrewPython,
    isMonolithPy,
    isMSYS2MingwPython,
    isPythonBuildStandalonePython,
)
from nuitka.PythonVersions import getSystemPrefixPath
from nuitka.Tracing import general, inclusion_logger
from nuitka.utils.Execution import executeToolChecked
from nuitka.utils.FileOperations import (
    areInSamePaths,
    getNormalizedPathJoin,
    isFilenameBelowPath,
    makePath,
    relpath,
)
from nuitka.utils.SharedLibraries import copyDllFile, setSharedLibraryRPATH
from nuitka.utils.Signing import addMacOSCodeSignature
from nuitka.utils.Timing import TimerReport
from nuitka.utils.Utils import (
    getOS,
    isDebianBasedLinux,
    isMacOS,
    isPosixWindows,
    isRPathUsingPlatform,
    isWin32Windows,
)

from .DllDependenciesMacOS import (
    detectBinaryPathDLLsMacOS,
    fixupBinaryDLLPathsMacOS,
)
from .DllDependenciesPosix import detectBinaryPathDLLsPosix
from .DllDependenciesWin32 import (
    detectBinaryPathDLLsWin32,
    shallIncludeVCRedistDLL,
)
from .IncludedEntryPoints import (
    addIncludedEntryPoint,
    getIncludedExtensionModule,
    getStandaloneEntryPointForSourceFile,
    makeDllEntryPoint,
)
from .MacOSApp import createEntitlementsInfoFile


def checkFreezingModuleSet():
    """Check the module set for troubles.

    Typically Linux OS specific packages must be avoided, e.g. Debian packaging
    does make sure the packages will not run on other OSes.
    """
    # Cyclic dependency
    from nuitka import ModuleRegistry

    problem_modules = OrderedSet()

    if isDebianBasedLinux():
        message = "Standalone with Python package from Debian installation may not be working."
        mnemonic = "debian-dist-packages"

        def checkModulePath(module):
            module_filename = module.getCompileTimeFilename()
            module_filename_parts = module_filename.split("/")

            if (
                "dist-packages" in module_filename_parts
                and "local" not in module_filename_parts
            ):
                module_name = module.getFullName()

                package_name = module_name.getTopLevelPackageName()

                if package_name is not None:
                    problem_modules.add(package_name)
                else:
                    problem_modules.add(module_name)

    else:
        checkModulePath = None
        message = None
        mnemonic = None

    # We intend for other platforms to join, e.g. Fedora, etc. but currently
    # only Debian is done.
    if checkModulePath is not None:
        for module in ModuleRegistry.getDoneModules():
            if not module.getFullName().isFakeModuleName():
                checkModulePath(module)

    if problem_modules:
        general.info("Using Debian packages for '%s'." % ",".join(problem_modules))
        general.warning(message=message, mnemonic=mnemonic)


def _detectBinaryDLLs(
    is_main_executable,
    source_dir,
    original_filename,
    binary_filename,
    package_name,
    use_path,
    use_cache,
    update_cache,
):
    """Detect the DLLs used by a binary.

    Using "ldd" (Linux), "depends.exe" (Windows), or
    "otool" (macOS) the list of used DLLs is retrieved.
    """

    if is_main_executable and isMonolithPy():
        return OrderedSet()
    elif (
        getOS() in ("Linux", "NetBSD", "FreeBSD", "OpenBSD", "AIX") or isPosixWindows()
    ):
        return detectBinaryPathDLLsPosix(
            dll_filename=original_filename,
            package_name=package_name,
            original_dir=os.path.dirname(original_filename),
        )
    elif isWin32Windows():
        with TimerReport(
            logger=inclusion_logger,
            message="Detecting dependencies for %s took %%.2f seconds"
            % binary_filename,
            decider=isShowProgress,
        ):
            return detectBinaryPathDLLsWin32(
                is_main_executable=is_main_executable,
                source_dir=source_dir,
                original_dir=os.path.dirname(original_filename),
                binary_filename=binary_filename,
                package_name=package_name,
                use_path=use_path,
                use_cache=use_cache,
                update_cache=update_cache,
            )
    elif isMacOS():
        return detectBinaryPathDLLsMacOS(
            original_dir=os.path.dirname(original_filename),
            binary_filename=original_filename,
            package_name=package_name,
            keep_unresolved=False,
            recursive=True,
        )
    else:
        # Support your platform above, for many platforms the POSIX branch will
        # be working and its parser for "ldd" just need fixes for specialties
        # maybe.
        raise NuitkaNotYetSupported(
            "DLL dependency detection for '%s', please help adding it." % getOS()
        )


def copyDllsUsed(dist_dir, standalone_entry_points):
    # Only do ones not ignored
    copy_standalone_entry_points = [
        standalone_entry_point
        for standalone_entry_point in standalone_entry_points[1:]
        if not standalone_entry_point.kind.endswith("_ignored")
        if "copy" in standalone_entry_point.tags
    ]
    main_standalone_entry_point = standalone_entry_points[0]

    if isMacOS():
        fixupBinaryDLLPathsMacOS(
            binary_filename=getNormalizedPathJoin(
                dist_dir, main_standalone_entry_point.dest_path
            ),
            package_name=main_standalone_entry_point.package_name,
            original_location=main_standalone_entry_point.source_path,
            standalone_entry_points=standalone_entry_points,
        )

    # After dependency detection, we can change the RPATH for main binary.
    if isRPathUsingPlatform():
        setSharedLibraryRPATH(
            getNormalizedPathJoin(dist_dir, standalone_entry_points[0].dest_path),
            "$ORIGIN",
        )

    setupProgressBar(
        stage="Copying used DLLs",
        unit="DLL",
        total=len(copy_standalone_entry_points),
    )

    for standalone_entry_point in copy_standalone_entry_points:
        reportProgressBar(standalone_entry_point.dest_path)

        copyDllFile(
            source_path=standalone_entry_point.source_path,
            dist_dir=dist_dir,
            dest_path=standalone_entry_point.dest_path,
            executable=standalone_entry_point.executable,
            other_entry_points=copy_standalone_entry_points,
        )

        if isMacOS():
            fixupBinaryDLLPathsMacOS(
                binary_filename=getNormalizedPathJoin(
                    dist_dir, standalone_entry_point.dest_path
                ),
                package_name=standalone_entry_point.package_name,
                original_location=standalone_entry_point.source_path,
                standalone_entry_points=standalone_entry_points,
            )

    closeProgressBar()

    onCopiedDLLs(
        dist_dir=dist_dir,
        standalone_entry_points=copy_standalone_entry_points,
    )

    return main_standalone_entry_point, copy_standalone_entry_points


def signDistributionMacOS(
    dist_dir, data_file_paths, main_standalone_entry_point, copy_standalone_entry_points
):
    # Make all top level directories symlinks for signing issues with MacOS
    # bundles. This is complex, because we also need to handle the need to
    # symlink and track information. pylint: disable=too-many-locals

    # spell-checker: ignore xattr
    executeToolChecked(
        logger=inclusion_logger,
        command=("/usr/bin/xattr", "-cr", dist_dir),
        absence_message="needs 'xattr' to remove extended attributes",
    )

    translations = OrderedSet()
    symlinks = OrderedSet()

    def _translatePath(path):
        for translation in translations:
            path = path.replace(translation[0] + "/", translation[1] + "/", 1)

        return path

    if shallCreateAppBundle():
        app_path = getNormalizedPathJoin(dist_dir, "..", "..")
        resources_dir = getNormalizedPathJoin(dist_dir, "..", "Resources")

        def getNotSignableDirectoryPart(filename):
            result = []

            for part in os.path.dirname(filename).split("/"):
                result.append(part)
                if "." in part:
                    return "/".join(result)

            return None

        for data_file_path in sorted(data_file_paths, key=len):
            data_file_path = _translatePath(data_file_path)

            inside_path = relpath(data_file_path, start=app_path)

            if not inside_path.startswith("Contents/MacOS"):
                continue

            not_signable_part = getNotSignableDirectoryPart(inside_path)
            if not_signable_part is None:
                continue

            filename = not_signable_part[len("Contents/MacOS/") :]
            not_signable_path = getNormalizedPathJoin(app_path, not_signable_part)
            resources_path = getNormalizedPathJoin(resources_dir, filename)

            if resources_path in symlinks:
                continue

            makePath(os.path.dirname(resources_path))
            os.rename(not_signable_path, resources_path)

            symlink_target = getNormalizedPathJoin("..", "Resources", filename)
            for _i in range(filename.count("/")):
                symlink_target = getNormalizedPathJoin("..", symlink_target)

            os.symlink(symlink_target, not_signable_path)

            symlinks.add(resources_path)
            translations.add((not_signable_path, resources_path))

    addMacOSCodeSignature(
        filenames=[
            _translatePath(filename)
            for filename in [
                getNormalizedPathJoin(dist_dir, standalone_entry_point.dest_path)
                for standalone_entry_point in [main_standalone_entry_point]
                + copy_standalone_entry_points
            ]
            + data_file_paths
        ],
        entitlements_filename=createEntitlementsInfoFile(),
    )


_excluded_system_dlls = set()


def _reduceToPythonPath(used_dll_paths):
    """Remove DLLs outside of python path unless they are found in the MSVC Redist folder."""
    inside_paths = getPythonUnpackedSearchPath()

    if isAnacondaPython():
        inside_paths.insert(0, getSystemPrefixPath())

    if isMacOS() and (isCPythonOfficialPackage() or isPythonBuildStandalonePython()):
        inside_paths.insert(0, getSystemPrefixPath())

    if isHomebrewPython():
        inside_paths.insert(0, getHomebrewInstallPath())

    if isMSYS2MingwPython():
        inside_paths.insert(0, getNormalizedPathJoin(getSystemPrefixPath(), "bin"))

    def decideInside(dll_filename):
        return any(
            isFilenameBelowPath(path=inside_path, filename=dll_filename)
            for inside_path in inside_paths
        )

    def decideMsvcRedistDll(dll_filename):
        return isWin32Windows() and shallIncludeVCRedistDLL(dll_filename)

    kept_used_dll_paths = OrderedSet()
    removed_dll_paths = OrderedSet()

    for dll_filename in used_dll_paths:
        if decideInside(dll_filename) or decideMsvcRedistDll(dll_filename):
            kept_used_dll_paths.add(dll_filename)
        else:
            if dll_filename not in _excluded_system_dlls:
                _excluded_system_dlls.add(dll_filename)
                inclusion_logger.debug("Not including system DLL '%s'" % dll_filename)

            removed_dll_paths.add(dll_filename)

    return kept_used_dll_paths, removed_dll_paths


_removed_dll_usages = {}


def getRemovedUsedDllsInfo():
    return _removed_dll_usages.items()


def _detectUsedDLLs(standalone_entry_point, source_dir):
    # TODO: We are handling a bunch of cases here, but also some special cases,
    # that should live elsewhere.
    # pylint: disable=too-many-branches,too-many-locals

    if standalone_entry_point.module_name is not None:
        # For Linux Pythons, there can be DLLs to pick up from the system.
        if (
            not isWin32Windows()
            and not isMacOS()
            and isStandardLibraryPath(standalone_entry_point.source_path)
        ):
            allow_outside_dependencies = True
        else:
            allow_outside_dependencies = decideAllowOutsideDependencies(
                standalone_entry_point.module_name
            )
    else:
        allow_outside_dependencies = False

    binary_filename = standalone_entry_point.source_path
    try:
        used_dll_paths = _detectBinaryDLLs(
            is_main_executable=standalone_entry_point.kind == "executable",
            source_dir=source_dir,
            original_filename=standalone_entry_point.source_path,
            binary_filename=binary_filename,
            package_name=standalone_entry_point.package_name,
            use_path=allow_outside_dependencies,
            use_cache=not shallNotUseDependsExeCachedResults(),
            update_cache=not shallNotStoreDependsExeCachedResults(),
        )
    except NuitkaForbiddenDLLEncounter:
        inclusion_logger.info(
            "Not including due to forbidden DLL '%s'." % binary_filename
        )
    except (RuntimeError, Exception) as e:
        inclusion_logger.warning(
            """\
Error, cannot detect used DLLs for DLL '%s' in package '%s' due to: %s"""
            % (binary_filename, standalone_entry_point.package_name, str(e))
        )

        raise
    else:
        # Plugins generally decide if they allow dependencies from the outside
        # based on the package name.

        if standalone_entry_point.module_name is not None and used_dll_paths:
            module_name, _module_filename, _kind, finding = locateModule(
                standalone_entry_point.module_name, parent_package=None, level=0
            )

            # Make sure we are not surprised here.
            assert module_name == standalone_entry_point.module_name, (
                module_name,
                "!=",
                standalone_entry_point.module_name,
            )
            assert finding == "absolute", standalone_entry_point.module_name

            if not allow_outside_dependencies:
                used_dll_paths, removed_dll_paths = _reduceToPythonPath(
                    used_dll_paths=used_dll_paths
                )

                if removed_dll_paths:
                    _removed_dll_usages[standalone_entry_point] = (
                        "no outside dependencies allowed for '%s'"
                        % standalone_entry_point.module_name,
                        removed_dll_paths,
                    )

        # Lets be sure of this
        assert type(used_dll_paths) is OrderedSet, type(used_dll_paths)

        # Allow plugins can prevent inclusion, this may discard things from
        # used_dlls through its return value.
        removed_dlls = removeDllDependencies(
            dll_filename=binary_filename, dll_filenames=OrderedSet(used_dll_paths)
        )
        used_dll_paths = used_dll_paths - removed_dlls

        for used_dll_path in used_dll_paths:
            extension_standalone_entry_point = getIncludedExtensionModule(used_dll_path)
            if extension_standalone_entry_point is not None:
                # Sometimes an extension module is used like a DLL, make sure to
                # remove it as a DLL then, there is no value in keeping those. Need
                # to keep it's destination path from that extension module then.
                dest_path = extension_standalone_entry_point.dest_path
            elif (
                standalone_entry_point.package_name is not None
                and standalone_entry_point.package_name.hasOneOfNamespaces(
                    "openvino",
                    "av",
                )
                and areInSamePaths(binary_filename, used_dll_path)
            ):
                # TODO: If used by a DLL from the same folder, put it there,
                # otherwise top level, but for now this is limited to a few cases
                # where required that way (openvino) or known to be good only (av),
                # because it broke other things. spell-checker: ignore openvino

                dest_path = getNormalizedPathJoin(
                    os.path.dirname(standalone_entry_point.dest_path),
                    os.path.basename(used_dll_path),
                )
            else:
                existing_entry_point = getStandaloneEntryPointForSourceFile(
                    source_path=used_dll_path,
                    package_name=standalone_entry_point.package_name,
                )

                if existing_entry_point is None:
                    dest_path = os.path.basename(used_dll_path)
                else:
                    dest_path = existing_entry_point.dest_path

            dll_entry_point = makeDllEntryPoint(
                logger=inclusion_logger,
                source_path=used_dll_path,
                dest_path=dest_path,
                module_name=standalone_entry_point.module_name,
                package_name=standalone_entry_point.package_name,
                reason="Used by '%s'" % standalone_entry_point.dest_path,
            )

            addIncludedEntryPoint(dll_entry_point)


def detectUsedDLLs(standalone_entry_points, source_dir):
    setupProgressBar(
        stage="Detecting used DLLs",
        unit="DLL",
        total=len(standalone_entry_points),
    )

    for standalone_entry_point in standalone_entry_points:
        reportProgressBar(standalone_entry_point.dest_path)

        _detectUsedDLLs(
            standalone_entry_point=standalone_entry_point, source_dir=source_dir
        )

    closeProgressBar()


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
